/*
 * Created on Jul 17, 2003
 */
package AcmeStudio.remote;

import java.rmi.RemoteException;
import java.util.List;

import AcmeStudio.remote.exception.ElementNotFoundException;
import AcmeStudio.remote.exception.TypeNotFoundException;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * This interface allows the client to manipulate an AcmeDesign in AcmeStudio.
 * The manipulation is limited to operations involving instances, that is
 * the AcmeElements, and cannot be applied to type-level objects.  For example,
 * a client cannot make changes to an AcmeFamily by adding types.<br>
 * <br>
 * The general usage of this interface is the following:<ol>
 * <li> Get a handle to an AcmeDesign by calling setupDesign (on interface
 * IAcmeStudio); this handle provides the access methods to
 * manipulate the AcmeDesign.
 * <li> Obtain the name of the only system in the AcmeDesign; this system name
 * is necessary for some of the methods
 * <li> Call any number of add/remove methods to instantiate or eliminate
 * components, connectors, ports, roles, attachments, bindings, or properties.
 * <li> Call assign/unassign methods to assign additional types to instances.
 * <li> Call set methods to alter values 
 * </ol>
 */
public interface IAcmeStudioDesign extends IAcmeStudioRemoteControl {

    public void close () throws RemoteException;

    /**
     * Returns the name of the AcmeSystem in this AcmeDesign.
     * @return String  Qualified name of the AcmeSystem
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find a specified AcmeElement
     */
    public String getSystem ()
    throws RemoteException, ElementNotFoundException;

    /**
     * Retrieves the name of the system within the named AcmeRepresentation. 
     * @param repName  name of the AcmeRepresentation
     * @return String  Qualified name of the system inside the AcmeRepresentation
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find a specified AcmeElement
     */
    public String getSystem (String repName)
    throws RemoteException, ElementNotFoundException;

    /**
     * Retrieves all the ports in the component named <code>compName</code>,
     * equivalent to invoking <code>getPortsOfType(dID, compName, null)</code>. 
     * @param compName  Name of the AcmeComponent whose port to search
     * @return List  List of qualified name of the ports, or null if there 
     * is none.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or component.
     */
    public List getPorts (String compName)
    throws RemoteException, ElementNotFoundException;

    /**
     * Retrieves the ports of the type specified in the component named
     * <code>compName</code>. 
     * @param compName  Name of the AcmeComponent whose port to search
     * @param type  Type of port to search
     * @return List  List of qualified name of the ports of the specified
     * type, or null if there is none.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or component.
     * @throws TypeNotFoundException  Unable to find specified port type
     */
    public List getPortsOfType (String compName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    /**
     * Retrieves the roles in the connector named <code>connName</code>,
     * equivalent to invoking <code>getRolesOfType(dID, compName, null)</code>.
     * @param connName  Name of the AcmeConnector whose role to search
     * @return List  List of qualified name of the roles, or null if there
     * is none.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or connector.
     */
    public List getRoles (String connName)
    throws RemoteException, ElementNotFoundException;

    /**
     * Retrieves the roles of the type specified in the connector named
     * <code>connName</code>.
     * @param connName  Name of the AcmeConnector whose role to search
     * @param type  Type of role to search
     * @return List  List of qualified name of the roles of the specified
     * type, or null if there is none.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or connector.
     * @throws TypeNotFoundException  Unable to find specified role type
     */
    public List getRolesOfType (String connName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    /**
     * Returns whether the named port is available, i.e., has no role attached.
     * @param portName  Qualified name of the port to check.
     * @return boolean  Indicates whether port is available.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or port.
     */
    public boolean isPortAvailable (String portName)
    throws RemoteException, ElementNotFoundException;

    /**
     * Returns whether the named role is available, i.e., has no port attached.
     * @param roleName  Qualified name of the role to check
     * @return boolean  Indicates whether role is availble.
     * @throws RemoteException
     * @throws ElementNotFoundException  Unable to find design or role.
     */
    public boolean isRoleAvailable (String roleName)
    throws RemoteException, ElementNotFoundException;


    public String addComponent (String sysName, String compName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addConnector (String sysName, String connName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addPort (String compName, String portName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addRole (String connName, String roleName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addAttachment (String sysName, String portName, String roleName)
    throws RemoteException, ElementNotFoundException;

    public String addRepresentation (String elementName, String name, String fam)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addBinding (String extPort, String intPort)
    throws RemoteException, ElementNotFoundException;

    public String addGauge (String elemName, String gaugeName, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public void removeComponent (String sysName, String compName)
    throws RemoteException, ElementNotFoundException;

    public void removeConnector (String sysName, String connName)
    throws RemoteException, ElementNotFoundException;

    public void removePort (String portName)
    throws RemoteException, ElementNotFoundException;

    public void removeRole (String roleName)
    throws RemoteException, ElementNotFoundException;

    public void removeAttachment (String portName, String roleName)
    throws RemoteException, ElementNotFoundException;

    public void removeRepresentation (String repName)
    throws RemoteException, ElementNotFoundException;

    public void removeBinding (String extPort, String intPort)
    throws RemoteException, ElementNotFoundException;

    public void removeGauge (String gaugeName)
    throws RemoteException, ElementNotFoundException;

    public String attachGauge (String propName, String gaugeName, String mapping)
    throws RemoteException, ElementNotFoundException;

    public String detachGauge (String propName, String gaugeName)
    throws RemoteException, ElementNotFoundException;

    /**
     * Sets the specified property of the specified element to a new value,
     * adding the property if it doesn't already exist.
     * @param dID  AcmeDesign unique ID
     * @param elementName  Qualified name of the AcmeElement in which to set property
     * @param propName  Name of property (qualified or relative) to set
     * @param type  Type of property, necessary to determine how to parse value
     * @param val  New value to which to set property
     * @return String  Previous value of property, if applicable 
     * @throws RemoteException
     * @throws ElementNotFoundException  Cannot find element or property
     * @throws TypeNotFoundException  Cannot find the type of the property specified
     */
    public String setProperty (String elementName, String propName, String type, String val)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public String addArmaniRule (String elementName, String rule)
    throws RemoteException, ElementNotFoundException;

    public void removeProperty (String elementName, String propName)
    throws RemoteException, ElementNotFoundException;

    public void removeArmaniRule (String elementName, String rule)
    throws RemoteException, ElementNotFoundException;


    public void assignFamily (String sys, String fam)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public void assignType (String target, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public void unassignFamily (String sys, String fam)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;

    public void unassignType (String target, String type)
    throws RemoteException, ElementNotFoundException, TypeNotFoundException;


//  public String addSystem (String name, String fam)
//  throws RemoteException;
//
//  public void removeSystem (int dID) throws RemoteException;
//
//
// Type-level methods
//  public String addFamily (String family)
//  throws RemoteException;
//    
//  public String addComponentType (String family, String type)
//  throws RemoteException;
//
//  public String addConnectorType (String family, String type)
//  throws RemoteException;
//
//  public String addPortType (String family, String type)
//  throws RemoteException;
//
//  public String addRoleType (String family, String type)
//  throws RemoteException;
//
//  public String addPropertyType (String family, String type)
//  throws RemoteException;
//
//  public void removeFamily (String family)
//  throws RemoteException;
//
//  public void removeComponentType (String family, String type)
//  throws RemoteException;
//
//  public void removeConnectorType (String family, String type)
//  throws RemoteException;
//
//  public void removePortType (String family, String type)
//  throws RemoteException;
//
//  public void removeRoleType (String family, String type)
//  throws RemoteException;
//
//  public void removePropertyType (String family, String type)
//  throws RemoteException;
//
}
