/*
 * Created on Jul 17, 2003
 */
package AcmeStudio.remote.impl;

import java.io.IOException;
import java.io.PipedWriter;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.jface.util.OpenStrategy;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.remote.IAcmeStudio;
import AcmeStudio.remote.IAcmeStudioDesign;
import AcmeStudio.remote.exception.SetupDesignException;
import AcmeStudio.remote.rmi.RemoteDesignUnicast;
import AcmeStudio.remote.utils.SetupRemote;
import AcmeStudio.ui.editors.diagram.figures.SystemView;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeOpenedDesignRecord;
import acmeLib.elements.AcmeSystem;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 * 
 * The class provides remote control access to AcmeStudio.  It serves remote
 * access requests from clients via RMI, and the methods correspond to
 * architecture change operations.<br>
 * <br>
 * This class is designed as a Singleton class so both the RemoteObject and 
 * AcmeStudio side can interact with the same instance.
 */
public class RemoteControlServer implements IAcmeStudio {

    /** Amount of time to sleep each cycle when waiting for async */
    protected static final int WAIT_ASYNC_TIME = 400;  // ms  
    /** Amount of time to pause to let pipe read */
    protected static final int WAIT_PIPE_TIME = 200;  // ms  

    /** Singleton server instance */
    private static RemoteControlServer _server = null;

    /** SWT Display used to interact with AcmeStudio */
    private Display _display = null;
    /** Active workbench page */
    private IWorkbenchPage _page = null;
    /** Channel for feedback to AcmeStudio Remote Control View */
    private PipedWriter _pipe = null;

    /** Flag indicating whether async execution is in progress */
    private boolean _asyncInProgress = false;

    /** Stores a map of AcmeDesign to IAcmeStudioDesign objects */
    private Map _map = null;

    /** Information used to access AcmeStudio */
    private ASInfo _info = null;

    /**
     * Returns the singleton instance of this server, lazy init if necessary.
     * @return RemoteControlServer
     */    
    public static RemoteControlServer getInstance () {
        if (_server == null) {
            _server = new RemoteControlServer();
        }
        return _server;
    }

    /**
     * Constructs an instance of RemoteControlServer to serve requests.
     */
    protected RemoteControlServer() {
        super();
        
        _display = Display.getDefault();
        _display.asyncExec(new Runnable() {
            public void run() {
                _page = AcmeStudioPlugin.getActiveWindow().getActivePage();
            }
        });

        _map = new HashMap();
    }

    /**
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        if (_pipe != null) {
            try {
                _pipe.close();
            } catch (IOException e) {
                // ignore
            }
            _pipe = null;
        }

        super.finalize();
    }

    /**
     * Returns the output end of the communication pipe, lazy init.
     * @return PipedWriter  the output pipe to which activity of this server
     * is reported.
     */
    public PipedWriter getPipe () {
        if (_pipe == null) {
            _pipe = new PipedWriter();
        } else {
            try {  // test for problems
                _pipe.flush();
            } catch (IOException e) {  // re-construct writer
                try {
                    _pipe.close();
                } catch (IOException e2) {
                    // ignore
                }
                _pipe = new PipedWriter();
            }
        }
        return _pipe;
    }

    ////
    // Remote Interface IAcmeStudioRemoteControl methods
    ////

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioRemoteControl#serverReady()
     */
    public boolean serverReady () throws RemoteException {
        return _page != null;
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioRemoteControl#setupSystem(String, String)
     */
    public IAcmeStudioDesign setupDesign (String fileName, String projName)
        throws RemoteException {
//        int retID = 0;

        IProject proj = null;
        if (projName != null) {  // first fetch the project
            pipe("Received call to setup design for file '" + fileName
                + "' under project '" + projName + "'.");
            proj = AcmeStudioPlugin.getRoot().getProject(projName);
        }
        if (proj == null) {  // get currently active project
            pipe("X> Unable to find project named " + projName + "!");
            proj = ASContext.current().getActiveProject();
            if (proj == null) {  // fetch the first project in array
                pipe("X> Unable to find currently active project!");
                proj = AcmeStudioPlugin.getRoot().getProjects()[0];
                if (proj == null) {
                    pipe("X> Unable to find any project!!");
                    throw new SetupDesignException(SetupDesignException.PROJECT_NOT_FOUND, projName);
                }
            }
        }

        // find file
        final IFile file = (IFile) proj.findMember(fileName);
        if (file == null) {
            throw new SetupDesignException(SetupDesignException.FILE_NOT_FOUND, fileName);
        }

        final ASInfo info = new ASInfo();

        String absName = null;
        try {
            absName = file.getLocation().toFile().getCanonicalPath();
        } catch (IOException e) {
            pipe("X> " + e.toString());
            e.printStackTrace();
            absName = file.getLocation().toString();
        }
        info.setFileName(absName);

        pipe("=> Opening editor on file '" + absName + "'");
        setAsyncInProgress(true);
        try {
            _display.asyncExec(new Runnable() {
                public void run() {
                    try {
                        info.setEditPart(_page.openEditor(new org.eclipse.ui.part.FileEditorInput (file), "AcmeStudio.ui.editors.MultiPageAcmeEditor", OpenStrategy.activateOnOpen()));
                        // assume it's MPAE, get and store the SystemView
                        info.setSystemView((SystemView) info.mPAE()
                            .getDiagramEditor().getRootEditPart().getFigure());
                        // get reference to AcmeSystem opened & store it
                        AcmeOpenedDesignRecord rec =
                            AcmeDesign.getDesignRecord(info.fileName());
                        if (rec != null) {
                            AcmeDesign design = rec.getResolvedDesign();
                            if (design != null) {
                                info.setDesign(design);
                                AcmeSystem sys = (AcmeSystem) design.getSystems().getNextItem();
                                info.setSystem(sys);
                            }
                        }
                    } catch (PartInitException e) {
                        throw new RuntimeException(new SetupDesignException(
                            SetupDesignException.OPEN_FAILED, file.toString(), e));
                    } finally {
                        setAsyncInProgress(false);
                    }
                }
            });
        } catch (RuntimeException e) {
            if (e.getCause() instanceof SetupDesignException) {
                pipe("X> Open editor failed for file " + file.toString() + "!");
                throw (SetupDesignException) e.getCause();
            } else {
                throw e;
            }
        }
        waitAsync();

        // store a mapping of identifier to file
//        retID = (int)System.currentTimeMillis() + absName.hashCode();
//        _map.put(new Integer(retID), info);
//        return retID;

        RemoteDesign rDesign = new RemoteDesign(info);
        _map.put(info.design(), rDesign);

        Remote remote = new RemoteDesignUnicast(rDesign);
        String rmiName = SetupRemote.SEP + SetupRemote.SEP + SetupRemote.HOST
            + SetupRemote.PORT_SEP + SetupRemote.RMI_PORT + SetupRemote.SEP
            + remote.getClass().getName() + System.currentTimeMillis();
        rDesign.setRMI(rmiName);
        try {
			Naming.rebind(rmiName, remote);
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		return (IAcmeStudioDesign) RemoteDesignUnicast.toStub(remote);
    }

    protected void setInfo (ASInfo info) {
        _info = info;
    }

    protected ASInfo getInfo () {
        return _info;
    }


    /**
     * Writes message to pipe if pipe was initialized, and flush pipe.
     * @param msg  message to write to pipe.
     */
    static protected void pipe (String msg) {
        PipedWriter pipe = getInstance().getPipe();

        if (pipe == null) return;  // can't send output yet

        try {
            pipe.write(msg);
            pipe.write("\n");  // necessary for receiving side buffered reader to return
            pipe.flush();
            try {
                Thread.sleep(WAIT_PIPE_TIME);
            } catch (InterruptedException e) { /* ignore */ }
        } catch (IOException e) {
            // ignore it
        }
    }

    protected void setAsyncInProgress (boolean s) {
        _asyncInProgress = s;
    }

    protected void waitAsync () {
        while (_asyncInProgress) {
            try {
                Thread.sleep(WAIT_ASYNC_TIME);
            } catch (InterruptedException e) {
                // ignore
            }
        }
    }

	/**
	 * @return Display  Eclipse Display associated with the editor that edits
     * the AcmeDesign controlled by this server instance.
	 */
	protected Display getDisplay() {
		return _display;
	}

}
