/*
 * Created on Aug 3, 2003
 */
package AcmeStudio.remote.impl;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;
import org.eclipse.draw2d.Figure;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.GraphicalEditPart;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.model.*;
import AcmeStudio.prefs.GlobalVisPreferences;
import AcmeStudio.remote.IAcmeStudioDesign;
import AcmeStudio.remote.exception.ElementNotFoundException;
import AcmeStudio.remote.exception.TypeNotFoundException;
import AcmeStudio.ui.editors.diagram.actions.SetElementConstraintCommand;
import AcmeStudio.ui.editors.diagram.figures.AcmeVisElement;
import AcmeStudio.ui.editors.diagram.figures.SystemView;
import AcmeStudio.ui.editors.diagram.styles.Location;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * The class provides remote control access to an AcmeDesign in AcmeStudio.
 * It serves remote access requests from clients via RMI, and the methods
 * correspond to architecture change operations.<br>
 */
public class RemoteDesign
    extends RemoteControlServer
    implements IAcmeStudioDesign {

    /** Design name */
    private String dN = null;
    /** The RMI remote object name for this object */
    private String _rmiName = null;

	/**
	 * 
	 */
	public RemoteDesign (ASInfo info) {
		super();
        
        setInfo(info);
        dN = info.design().getQualifiedName();
	}

    public void setRMI (String rmiName) {
        _rmiName = rmiName;
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#close()
     */
    public void close () throws RemoteException {
        if (_rmiName == null) {
            // nothing to do?
        } else {
            try {
				Remote remote = Naming.lookup (_rmiName);
                Naming.unbind(_rmiName);
			} catch (MalformedURLException e) {
				e.printStackTrace();
			} catch (RemoteException e) {
				e.printStackTrace();
			} catch (NotBoundException e) {
				e.printStackTrace();
			}
            
        }
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getSystem(int)
     */
    public String getSystem () throws RemoteException {
        pipe("Received call to get System name in design '" + dN + "'");

        return getInfo().system().getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getSystem(int, java.lang.String)
     */
    public String getSystem (String repName) throws RemoteException {
        pipe("Received call to get System name in Representation '" + repName + "'");

        final ASInfo i = getInfo();
        AcmeRepresentation rep = findRep(i.design(), repName);
        return rep.getSystem().getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getPorts(int, java.lang.String)
     */
    public List getPorts (String compName)
        throws RemoteException {

        return getPortsOfType(compName, null);
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getPortsOfType(int, java.lang.String, java.lang.String)
     */
    public List getPortsOfType (String compName, String type)
        throws RemoteException {
        pipe("Received call to find Ports of type '" + type + "' in Component '"
            + compName + "'");
        List retPorts = new ArrayList();

        final ASInfo i = getInfo();
        // find the component
        final AcmeComponent comp = findComp(i.design(), compName);
        if (type == null) {  // return all the ports
            AcmeEnumeration enum = comp.getPorts();
            while (enum.moreItems()) {
                AcmePort port = (AcmePort) enum.getNextItem();
                retPorts.add(port.getQualifiedName());
            }
        } else {
            // find the port type referred to
            final AcmePortType portT = findPortType(i.system(), type);
            // find in component an available port of the port type
            AcmeEnumeration enum = comp.getPorts();
            OUTER: while (enum.moreItems()) {
                AcmePort port = (AcmePort) enum.getNextItem();
                AcmeEnumeration ptEnum = port.getDeclaredTypes();
                ptEnum.concatenate(port.getInstantiatedTypes());
                while (ptEnum.moreItems()) {
                    AcmeElementType pType = (AcmeElementType ) ptEnum.getNextItem();
                    if (pType.getTypeName().equalsIgnoreCase(type)
                        && ModelHelper.isAvailablePort(port)) {
                        // port is available AND of the right type
                        retPorts.add(port.getQualifiedName());
                        break OUTER;
                    }
                }
            }
        }
        return retPorts;
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getRoles(int, java.lang.String)
     */
    public List getRoles (String connName)
        throws RemoteException {

        return getRolesOfType(connName, null);
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#getRolesOfType(int, java.lang.String, java.lang.String)
     */
    public List getRolesOfType (String connName, String type)
        throws RemoteException {
        pipe("Received call to find Roles of type '" + type + "' in Connector '"
            + connName + "'");
        List retRoles = new ArrayList();

        final ASInfo i = getInfo();
        // find the connector
        final AcmeConnector conn = findConn(i.design(), connName);
        if (type == null) {  // return all the ports
            AcmeEnumeration enum = conn.getRoles();
            while (enum.moreItems()) {
                AcmeRole role = (AcmeRole) enum.getNextItem();
                retRoles.add(role.getQualifiedName());
            }
        } else {
            // find the role type referred to
            final AcmeRoleType roleT = findRoleType(i.system(), type);
            // find in connector an available role of the role type
            AcmeEnumeration enum = conn.getRoles();
            OUTER: while (enum.moreItems()) {
                AcmeRole role = (AcmeRole) enum.getNextItem();
                AcmeEnumeration rtEnum = role.getDeclaredTypes();
                rtEnum.concatenate(role.getInstantiatedTypes());
                while (rtEnum.moreItems()) {
                    AcmeElementType rType = (AcmeElementType ) rtEnum.getNextItem();
                    if (rType.getTypeName().equalsIgnoreCase(type)
                        && ModelHelper.isAvailableRole(role)) {
                        // port is available AND of the right type
                        retRoles.add(role.getQualifiedName());
                        break OUTER;
                    }
                }
            }
        }
        return retRoles;
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#isPortAvailable(int, java.lang.String)
     */
    public boolean isPortAvailable (String portName)
        throws RemoteException {
        pipe("Received call to determine if Port '" + portName + "' in available'");

        final ASInfo i = getInfo();
        // find the port
        final AcmePort port = findPort(i.design(), portName);
        return ModelHelper.isAvailablePort(port);
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#isRoleAvailable(int, java.lang.String)
     */
    public boolean isRoleAvailable (String roleName)
        throws RemoteException {
        pipe("Received call to determine if Role '" + roleName + "' in available'");

        final ASInfo i = getInfo();
        // find the role
        final AcmeRole role = findRole(i.design(), roleName);
        return ModelHelper.isAvailableRole(role);
    }


    // Instance-level methods

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addComponent(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addComponent (String sysName, String compName, String type)
        throws RemoteException {
        pipe("Received call to add Component '" + compName + "' of type '"
            + type + "' to sys '" + sysName + "'");

        final ASInfo i = getInfo();
        // find the system
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the type referred to
        final AcmeComponentType compT = findCompType(sys, type);
        // create the component
        final String uName = Helper.getUniqueName(sys, compName);
        AcmeComponent comp = new AcmeComponent(uName, compT);
        comp.addInstantiatedType(compT);
        comp.instantiateElements();
        createElement(i, sys, comp);
        pipe("=> Added '" + comp.getTypeSignature() + "' to '"
            + sys.getTypeSignature() + "'");

        return comp.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addConnector(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addConnector (String sysName, String connName, String type)
        throws RemoteException {
        pipe("Received call to add Connector '" + connName + "' of type '"
            + type + "' to sys '" + sysName + "'");

        final ASInfo i = getInfo();
        // find the system
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the type referred to
        final AcmeConnectorType connT = findConnType(sys, type);
        // create the connector
        final String uName = Helper.getUniqueName(sys, connName);
        AcmeConnector conn = new AcmeConnector(uName, connT);
        conn.addInstantiatedType(connT);
        conn.instantiateElements();
        createElement(i, sys, conn);
        pipe("=> Added '" + conn.getTypeSignature() + "' to '"
            + sys.getTypeSignature() + "'");
        
        return conn.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addPort(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addPort (String compName, String portName, String type)
        throws RemoteException {
        pipe("Received call to add Port '" + portName + "' of type '" + type
            + "' to Component '" + compName + "' of sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the component referred to
        final AcmeComponent comp = findComp(i.design(), compName);
        // find the type referred to
        final AcmePortType portT = findPortType(i.system(), type);
        // create the port
        final String uName = Helper.getUniqueName(comp, portName);
        AcmePort port = new AcmePort(uName, portT);
        port.addInstantiatedType(portT);
        port.instantiateElements();
        createElement(i, comp, port);
        pipe("=> Added '" + port.getTypeSignature() + "' to '"
            + comp.getTypeSignature() + "'");

        return port.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addRole(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addRole (String connName, String roleName, String type)
        throws RemoteException {
        pipe("Received call to add Role '" + roleName + "' of type '" + type
            + "' to Connector'" + connName + "' of sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the connector referred to
        final AcmeConnector conn = findConn(i.design(), connName);
        // find the type referred to
        final AcmeRoleType roleT = findRoleType(i.system(), type);
        // create the role
        final String uName = Helper.getUniqueName(conn, roleName);
        AcmeRole role = new AcmeRole(uName, roleT);
        role.addInstantiatedType(roleT);
        role.instantiateElements();
        createElement(i, conn, role);
        pipe("=> Added '" + role.getTypeSignature() + "' to '"
            + conn.getTypeSignature() + "'");

        return role.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addAttachment(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addAttachment (String sysName, String portName, String roleName)
        throws RemoteException {
        pipe("Received call to attach port '" + portName + "' to role '"
            + roleName + "' in sys '" + sysName + "'");

        final ASInfo i = getInfo();
        // find the system
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the port referred to
        final AcmePort port = findPort(i.design(), portName);
        // find the role referred to
        final AcmeRole role = findRole(i.design(), roleName);
        // create the attachment
        AcmeAttachment attach = new AcmeAttachment(port, role);
        attach.setParent(sys);
        createAttachment(i, sys, attach);
        pipe("=> Added '" + attach.getTypeSignature() + "' to '"
            + sys.getTypeSignature() + "'");

        return attach.getQualifiedName();
    }

  
	/* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addRepresentation(int, java.lang.String, java.lang.String, java.lang.String)
     */
    public String addRepresentation (String elementName, String name, String fam)
        throws RemoteException {
        pipe("Received call to add Representation '" + name + "' of family '"
            + fam + "' to element '" + elementName + "' in sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the target element referred to
        final AcmeElement acmeE = findElement(i.design(), elementName);
        // find the family referred to
        final AcmeFamily famT = findFamily(i.design(), fam);
        final String uName = Helper.getUniqueName(acmeE, name);
        // create the descendant system for the representation
        AcmeSystem sys = new AcmeSystem(uName);
        sys.addDeclaredType(famT);
        sys.addInstantiatedType(famT);
        sys.instantiateElements();
        sys.setParentDesign(i.design());
        // create a bindings map
        AcmeBindingsMap bindMap = new AcmeBindingsMap();
        // now we can create the representation
        AcmeRepresentation rep = new AcmeRepresentation(sys, acmeE, bindMap, name);
        rep.instantiateElements();
        createRepresentation(i, acmeE, rep);
        pipe("=> Added '" + sys.getTypeSignature() + "' to '"
            + acmeE.getTypeSignature() + "'");

        return rep.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addBinding(int, java.lang.String, java.lang.String)
     */
    public String addBinding (String extName, String intName)
        throws RemoteException {
        pipe("Received call to bind external port '" + extName
            + "' to internal port '" + intName + "' in sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the external port referred to
        final AcmePort extPort = findPort(i.design(), extName);
        // find the internal port referred to
        final AcmePort intPort = findPort(i.design(), intName);
        // create the binding
        AcmeBinding binding = new AcmeBinding(extPort, intPort);
        binding.setParentRep(intPort.getParentRepresentation());
        try {
            binding.setName("Binding_" + extPort.getName() + "_" + intPort.getName());
        } catch (AcmeIllegalNameException e) {
            pipe("X> Setting binding name failed!");
        }
        createElement(i, null, binding);
        pipe("=> Added '" + binding.getTypeSignature() + "' to '"
            + i.system().getTypeSignature() + "'");

        return binding.getQualifiedName();
    }

	/* (non-Javadoc)
	 * @see AcmeStudio.remote.IAcmeStudioDesign#addGauge(java.lang.String, java.lang.String, java.lang.String)
	 */
	public String addGauge(String elemName, String gaugeName, String type)
			throws RemoteException, ElementNotFoundException,
			TypeNotFoundException {
		// TODO Auto-generated addGauge
        pipe("Received call to add Gauge '" + gaugeName + "' of type '" + type + "' to element '"
                + elemName + "'");
		return null;
	}

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeComponent(int, java.lang.String, java.lang.String)
     */
    public void removeComponent (String sysName, String compName) throws RemoteException {
        pipe("Received call to remove Component '" + compName + "' from sys '"
            + sysName + "'");

        final ASInfo i = getInfo();
        // find the system
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the component
        final AcmeComponent comp = findComp(i.design(), compName);
        deleteElement(i, sys, comp);
        pipe("=> Removed '" + comp.getTypeSignature() + "' from '"
            + sys.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeConnector(int, java.lang.String, java.lang.String)
     */
    public void removeConnector (String sysName, String connName) throws RemoteException {
        pipe("Received call to remove Connector '" + connName + "' from sys '"
            + sysName + "'");

        final ASInfo i = getInfo();
        // find the system
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the connector
        final AcmeConnector conn = findConn(i.design(), connName);
        deleteElement(i, sys, conn);
        pipe("=> Removed '" + conn.getTypeSignature() + "' from '"
            + sys.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removePort(int, java.lang.String)
     */
    public void removePort (String portName) throws RemoteException {
        pipe("Received call to remove Port '" + portName + "' from sys '"
            + dN + "'");

        final ASInfo i = getInfo();
        // find the port
        final AcmePort port = findPort(i.design(), portName);
        deleteElement(i, port.getParent(), port);
        pipe("=> Removed '" + port.getTypeSignature() + "' from '"
            + i.system().getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeRole(int, java.lang.String)
     */
    public void removeRole (String roleName) throws RemoteException {
        pipe("Received call to remove Role '" + roleName + "' from sys '"
            + dN + "'");

        final ASInfo i = getInfo();
        // find the role
        final AcmeRole role = findRole(i.design(), roleName);
        deleteElement(i, role.getParent(), role);
        pipe("=> Removed '" + role.getTypeSignature() + "' from '"
            + i.system().getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeAttachment(int, java.lang.String, java.lang.String)
     */
    public void removeAttachment (String portName, String roleName)
        throws RemoteException {
        pipe("Received call to detach port '" + portName + "' from role '"
            + roleName + "' in sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the port referred to
        final AcmePort port = findPort(i.design(), portName);
        // find the role referred to
        final AcmeRole role = findRole(i.design(), roleName);
        final AcmeAttachment attach = i.system().getAttachment(port, role);
        if (attach == null) {
            pipe("X> Unable to find AcmeAttachment between port " + portName
                + " & role " + roleName);
            throw new ElementNotFoundException(ElementNotFoundException.NO_ATTACH,
                "Between port " + portName + " & role " + roleName);
        }
        deleteElement(i, attach.getParent(), attach);
        pipe("=> Removed '" + attach.getTypeSignature() + "' from '"
            + i.system().getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeRepresentation(int, java.lang.String)
     */
    public void removeRepresentation (String name)
        throws RemoteException {
        pipe("Received call to remove Representation '" + name + "' from sys '"
            + dN + "'");

        final ASInfo i = getInfo();
        // find the system of the representation referred to
        final AcmeRepresentation rep = findRep(i.design(), name);
        deleteRepresentation(i, rep);
        pipe("=> Removed '" + name + "' from '"
            + i.system().getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeBinding(int, java.lang.String, java.lang.String)
     */
    public void removeBinding (String extName, String intName)
        throws RemoteException {
        pipe("Received call to unbind external port '" + extName
            + "' from internal port '" + intName + "' in sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the external port referred to
        final AcmePort extPort = findPort(i.design(), extName);
        // find the internal port referred to
        final AcmePort intPort = findPort(i.design(), intName);
        final AcmeRepresentation rep = intPort.getParentRepresentation();
        if (rep == null) {
            pipe("X> Unable to find AcmeRepresentation containing the interior port " + intName);
            throw new ElementNotFoundException(ElementNotFoundException.NO_REP, "Representation containing the interior port " + intName);
        }
        AcmeBindingsMap bm = (AcmeBindingsMap) rep.getAbstractionMap();
        if (bm == null) {
            pipe("X> No bindings map in the representation " + rep.getQualifiedName());
        } else {
            AcmeBinding binding = bm.getBinding(extPort, intPort);
            deleteElement(i, null, binding);
            pipe("=> Removed '" + binding.getTypeSignature()
                + "' from '" + i.system().getTypeSignature() + "'");
        }
    }

	/* (non-Javadoc)
	 * @see AcmeStudio.remote.IAcmeStudioDesign#removeGauge(java.lang.String)
	 */
	public void removeGauge(String gaugeName) throws RemoteException,
			ElementNotFoundException {
		// TODO Auto-generated removeGauge
        pipe("Received call to remove Gauge '" + gaugeName + "'");
	}

	/* (non-Javadoc)
	 * @see AcmeStudio.remote.IAcmeStudioDesign#attachGauge(java.lang.String, java.lang.String, java.lang.String)
	 */
	public String attachGauge(String propName, String gaugeName, String mapping)
			throws RemoteException, ElementNotFoundException {
		// TODO Auto-generated attachGauge
        pipe("Received call to attach Gauge '" + gaugeName + "' of mapping '" + mapping + "' to property '"
                + propName + "'");
		return null;
	}

	/* (non-Javadoc)
	 * @see AcmeStudio.remote.IAcmeStudioDesign#detachGauge(java.lang.String, java.lang.String)
	 */
	public String detachGauge(String propName, String gaugeName)
			throws RemoteException, ElementNotFoundException {
		// TODO Auto-generated detachGauge
        pipe("Received call to detach Gauge '" + gaugeName + "' from property '" + propName + "'");
		return null;
	}

	/* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#setPropety(int, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
     */
    public String setProperty (String elementName, String propName, String type, String val)
        throws RemoteException {
        pipe("Received call to set Property '" + propName + "' of type '" + type
            + "' in '" + elementName + "' to value '" + val + "'");

        final ASInfo i = getInfo();
        // find the element referred to
        final AcmeElement element = findElement(i.design(), elementName);
        // find the property referred to
        AcmeProperty prop = null;
        if (propName.startsWith ("vis-") && AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean(GlobalVisPreferences.KEEP_VIS_SEPARATE)) {
        	AcmeProperty tempProp = determinePropVal(i.system(), type, val);
        	try {
				tempProp.setName (propName);
			}
			catch (AcmeIllegalNameException e) {
				e.printStackTrace();
			}
        	setProperty (i, element, tempProp, tempProp.getValue ());
        	pipe ("=> Set property " + tempProp.getQualifiedName () + " to value '" + val + "'");
        	return propName;
        	
        }
        else { 
	        try {
	            prop = findProperty(i.design(), element, propName);
	            AcmeProperty tempProp = determinePropVal(i.system(), type, val);
	            setProperty(i, element, prop, tempProp.getValue());
	            pipe("=> Set property '" + prop.getQualifiedName()
	                + "' to new value '" + val + "'");
	        } catch (ElementNotFoundException e) {  // add it
	            AcmeProperty tempProp = determinePropVal(i.system(), type, val);
	            final String uName = Helper.getUniqueName(element, propName);
	            prop = new AcmeProperty(uName, tempProp.getType(), tempProp.getValue());
	            createProperty(i, element, prop);
	            pipe("=> Added new property '" + prop.getQualifiedName()
	                + "' of value '" + val + "' to '" + element.getTypeSignature() + "'");
	        }
        }
        return prop.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#addArmaniRule(int, java.lang.String, java.lang.String)
     */
    public String addArmaniRule (String elementName, String rule)
        throws RemoteException {
        // TODO RemoteControlServer.addArmaniRule
        pipe("Received call to add Armani Rule '" + rule + "' to element '"
            + elementName + "' in sys '" + dN + "'");

//        ExpressionNode node = ExpressionNode.
//        DesignRule rule = new DesignRule()
        return null;
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeProperty(int, java.lang.String)
     */
    public void removeProperty (String name, String propName) throws RemoteException {
        pipe("Received call to remove Property '" + propName + "' from '"
            + name + "' in sys '" + dN + "'");

        final ASInfo i = getInfo();
        // find the element referred to
        final AcmeElement element = findElement(i.design(), name);
        // find the property referred to
        final AcmeProperty prop = findProperty(i.design(), element, propName);
        deleteProperty(i, element, prop);
        pipe("=> Removed '" + prop.getQualifiedName() + "' from '"
            + element.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#removeArmaniRule(int, java.lang.String, java.lang.String)
     */
    public void removeArmaniRule(String elementName, String rule)
        throws RemoteException {
        // TODO RemoteControlServer.removeArmaniRule

    }



    // "Type"-assignment methods

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#assignFamily(int, java.lang.String, java.lang.String)
     */
    public void assignFamily (String sysName, String fam)
        throws RemoteException {
        pipe("Received call to assign Family '" + fam + "' to System '"
            + sysName + "' in design '" + dN + "'");

        final ASInfo i = getInfo();
        // find the target system referred to
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the family referred to
        final AcmeFamily famT = findFamily(i.design(), fam);
        // assign the family to the target system
        assignFamily(i, sys, famT);

        pipe("=> Assigned family '" + fam + "' to '" + sys.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#assignType(int, java.lang.String, java.lang.String)
     */
    public void assignType (String target, String type)
        throws RemoteException {
        pipe("Received call to assign type '" + type + "' to element '"
            + target + "' in design '" + dN + "'");

        final ASInfo i = getInfo();
        // find the target element referred to
        final AcmeElement acmeE = findElement(i.design(), target);
        // find the type referred to
        final AcmeElementType elementT = findElementType(i.system(), type);
        // assign the type to the target element
        assignType(i, acmeE, elementT);

        pipe("=> Assigned type '" + type + "' to '" + acmeE.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#unassignFamily(int, java.lang.String, java.lang.String)
     */
    public void unassignFamily (String sysName, String fam)
        throws RemoteException {
        pipe("Received call to unassign Family '" + fam + "' from System '"
            + sysName + "' in design '" + dN + "'");

        final ASInfo i = getInfo();
        // find the target system referred to
        final AcmeSystem sys = findSystem(i.design(), sysName);
        // find the family referred to
        final AcmeFamily famT = findFamily(i.design(), fam);
        // unassign the family from the target system
        assignFamily(i, sys, famT);

        pipe("=> Unassigned family '" + fam + "' from '"
            + sys.getTypeSignature() + "'");
    }

    /* (non-Javadoc)
     * @see AcmeStudio.remote.IAcmeStudioDesign#unassignType(int, java.lang.String, java.lang.String)
     */
    public void unassignType (String target, String type)
        throws RemoteException {
        pipe("Received call to unassign type '" + type + "' from element '"
            + target + "' in design '" + dN + "'");

        final ASInfo i = getInfo();
        // find the target element referred to
        final AcmeElement acmeE = findElement(i.design(), target);
        // find the type referred to
        final AcmeElementType elementT = findElementType(i.system(), type);
        // unassign the type from the target element
        unassignType(i, acmeE, elementT);

        pipe("=> Unassigned type '" + type + "' from '"
            + acmeE.getTypeSignature() + "'");
    }

    ////
    // Protected & private methods
    ////

    // Command execution methods

    /**
     * Asynchronously executes on AcmeStudio UI thread to create representation
     * on the parent element, using a RepresentationCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param parent  the parent AcmeElement to which to add the child
     * @param rep  the AcmeRepresentation to add to the parent
     */
    protected void createRepresentation (final ASInfo info, final AcmeElement parent, final AcmeRepresentation rep) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    RepresentationCommand repCmd =
                        new RepresentationCommand(RepresentationCommand.ADD_REP, rep, parent, this);
                    String label = "Add representation " + rep.getQualifiedName();
                    repCmd.setLabel(label);
                    repCmd.setDebugLabel(label);
                    repCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to delete representation
     * from the parent element, using a RepresentationCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param rep  the AcmeRepresentation to delete
     */
    protected void deleteRepresentation (final ASInfo info, final AcmeRepresentation rep) {
        // determine parent element
        final AcmeElement parent = rep.getParent();

        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    RepresentationCommand repCmd =
                        new RepresentationCommand(RepresentationCommand.REMOVE_REP, rep, parent, this);
                    String label = "Remove representation " + rep.getQualifiedName();
                    repCmd.setLabel(label);
                    repCmd.setDebugLabel(label);
                    repCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    
    
    /**
     * Asynchronously executes on AcmeStudio UI thread to create the child
     * element on the parent element, using a CreateElementCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param parent  the parent AcmeElement to which to add the child
     * @param child  the child AcmeElement to add to the parent
     */
    protected void createElement (final ASInfo info, final AcmeElement parent, final AcmeElement child) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    SystemView sysView = info.systemView();
             
                    Point location = new Point(0,0);
                    CreateElementCommand create =
                        new CreateElementCommand(parent, child, location, this);
                    Figure parentVis = (Figure )info.mPAE().getDiagramEditor().getFigure(parent);
                    if (parentVis instanceof AcmeVisElement) {
                    	create.setParentVis((AcmeVisElement )parentVis);
                    }
                    create.setSystemView(sysView);
                    Dimension size = Helper.getSize(child);
                    if (size != null) {
                        sysView.updateElement(child, new Rectangle(location, size));
                        sysView.computeChildLocations();
                        Location newL = null;
                        if ((newL = sysView.getCalculatedLocation(child)) != null) {
                            location = newL.getRectangle().getLocation();
                        }
                        create.setLocation(location);
                    }
                    String label = "Create " + child.getTypeSignature();
                    create.setLabel(label);
                    create.setDebugLabel(label);
                    create.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }
    
    /**
     * Asynchronously executes on AcmeStudio UI thread to add an attachment
     * to a system. There is no add attachment command, so have to do this inline
	 * @param i AcmeStudio info object to provide necessary support data
	 * @param sys the system into which the attachment is going	
	 * @param attach the attachment to add
	 */
	private void createAttachment(ASInfo i, final AcmeSystem sys, final AcmeAttachment attach) {
		setAsyncInProgress(true);
		getDisplay ().asyncExec(new Runnable () {
			
			public void run() {
				try {
					sys.addAttachment(attach);
					AcmeStudioChangeManager.globalChangeManager.onAddAttachment(sys, attach, this);
				}
				finally {
					setAsyncInProgress(false);
				}
				
			}
		});
		waitAsync();
	}


    /**
     * Asynchronously executes on AcmeStudio UI thread to delete the child
     * element from the parent element, using a DeleteElementCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param parent  the parent AcmeElement from which to delete the child
     * @param child  the child AcmeElement to delete from the parent
     */
    protected void deleteElement (final ASInfo info, final AcmeElement parent, final AcmeElement child) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    DeleteElementCommand delete =
                        new DeleteElementCommand(parent, child, this);
                    String label = "Delete " + child.getTypeSignature();
                    delete.setLabel(label);
                    delete.setDebugLabel(label);
                    delete.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to add property
     * to the element, using a PropertyCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param element  the AcmeElement to which to add the property
     * @param prop  the AcmeProperty to add to the element
     */
    protected void createProperty (final ASInfo info, final AcmeElement element, final AcmeProperty prop) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    PropertyCommand propCmd =
                        new PropertyCommand(PropertyCommand.ADD_PROP, element, prop, this);
                    String label = "Add property " + prop.getQualifiedName();
                    propCmd.setLabel(label);
                    propCmd.setDebugLabel(label);
                    propCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to set property
     * of the element to a new value, using a ChangePropertyCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param element  the AcmeElement to which to set the property value
     * @param prop  the AcmeProperty on which to set new value
     * @param val  the new AcmePropertyValue to which to set the property 
     */
    protected void setProperty (final ASInfo info, final AcmeElement element, final AcmeProperty prop, final AcmePropertyValue val) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                	if (isVisProp (prop) && AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean(GlobalVisPreferences.KEEP_VIS_SEPARATE)) {
                		SystemView sysView = info.systemView();
                		Point location = sysView.getElementLocation(element);
                		Dimension size = Helper.getSize (element);
                		if (prop.getName ().equals ("vis-x")) {
                			location.x = (int )((AcmeFloat )prop.getValue()).getValue();
                		}
                		else if (prop.getName ().equals ("vis-y")) {
                			location.y = (int )((AcmeFloat )prop.getValue ()).getValue ();
                		}
                		else if (prop.getName ().equals ("vis-width")) {
                			size.height = (int )((AcmeFloat )prop.getValue ()).getValue ();
                		}
                		else if (prop.getName ().equals ("vis-height")) {
                			size.width = (int )((AcmeFloat )prop.getValue ()).getValue ();
                		}
                		EditPart editPart = info.mPAE().getDiagramEditor().getEditPart(element);
                		if (editPart != null) {
                			SetElementConstraintCommand command = new SetElementConstraintCommand ((GraphicalEditPart )editPart);
                			command.setSystemView (sysView);
                			command.setElement (element);
                			command.setLocation (new Rectangle (location,size));
                			command.setLabel ("Change location/size (remote)");
                			command.setDebugLabel("Change location or size (done by remote control)");
                			command.execute ();
                		}
                		
                		
                	}
                	else {
	                    ChangePropertyCommand propCmd =
	                        new ChangePropertyCommand();
	                    // set up the necessary info for change property command
	                    propCmd.setTarget(element);
	                    propCmd.setProperty(prop);
	                    propCmd.setNewValue(val);
	                    String label = "Change property " + prop.getQualifiedName();
	                    propCmd.setLabel(label);
	                    propCmd.setDebugLabel(label);
	                    propCmd.execute(info.design());
                	}
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    private boolean isVisProp (AcmeProperty prop) {
    	return prop.getName ().startsWith ("vis-");
    }
    /**
     * Asynchronously executes on AcmeStudio UI thread to delete property
     * from the element, using a PropertyCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param element  the AcmeElement from which to delete the property
     * @param prop  the AcmeProperty to delete from the element
     */
    protected void deleteProperty (final ASInfo info, final AcmeElement element, final AcmeProperty prop) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    PropertyCommand propCmd =
                        new PropertyCommand(PropertyCommand.REMOVE_PROP, element, prop, this);
                    String label = "Remove property " + prop.getQualifiedName();
                    propCmd.setLabel(label);
                    propCmd.setDebugLabel(label);
                    propCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to assign specified
     * family to the system, using a TypeAssignmentCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param sys  the AcmeSystem to which to assign the family
     * @param fam  the AcmeFamily to assign to AcmeSystem
     */
    protected void assignFamily (final ASInfo info, final AcmeSystem sys, final AcmeFamily fam) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    TypeAssignmentCommand assignCmd =
                        new TypeAssignmentCommand(TypeAssignmentCommand.ASSIGN_FAMILY, sys, fam, this);
                    // set up the necessary info for type assignment command
                    String label = "Assign family " + fam.getName();
                    assignCmd.setLabel(label);
                    assignCmd.setDebugLabel(label);
                    assignCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to unassign specified
     * family from the system, using a TypeAssignmentCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param sys  the AcmeSystem from which to unassign the family
     * @param fam  the AcmeFamily to unassign from AcmeSystem
     */
    protected void unassignFamily (final ASInfo info, final AcmeSystem sys, final AcmeFamily fam) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    TypeAssignmentCommand assignCmd =
                        new TypeAssignmentCommand(TypeAssignmentCommand.UNASSIGN_FAMILY, sys, fam, this);
                    // set up the necessary info for type assignment command
                    String label = "Unassign family " + fam.getName();
                    assignCmd.setLabel(label);
                    assignCmd.setDebugLabel(label);
                    assignCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to assign specified type
     * to the element, using a TypeAssignmentCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param element  the AcmeElement to which to assign the type
     * @param type  the AcmeElementType to assign to AcmeElement
     */
    protected void assignType (final ASInfo info, final AcmeElement element, final AcmeElementType type) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    TypeAssignmentCommand assignCmd =
                        new TypeAssignmentCommand(TypeAssignmentCommand.ASSIGN_TYPE, element, type, true, this);
                    // set up the necessary info for type assignment command
                    String label = "Assign type " + type.getTypeName();
                    assignCmd.setLabel(label);
                    assignCmd.setDebugLabel(label);
                    assignCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    /**
     * Asynchronously executes on AcmeStudio UI thread to unassign specified
     * type from the element, using a TypeAssignmentCommand.
     * @param info  AcmeStudio info object to provide necessary support data
     * @param element  the AcmeElement from which to unassign the type
     * @param type  the AcmeElementType to unassign from AcmeElement
     */
    protected void unassignType (final ASInfo info, final AcmeElement element, final AcmeElementType type) {
        setAsyncInProgress(true);
        getDisplay().asyncExec(new Runnable() {
            public void run() {
                try {
                    TypeAssignmentCommand assignCmd =
                        new TypeAssignmentCommand(TypeAssignmentCommand.UNASSIGN_TYPE, element, type, true, this);
                    // set up the necessary info for type unassignment command
                    String label = "Unassign type " + type.getTypeName();
                    assignCmd.setLabel(label);
                    assignCmd.setDebugLabel(label);
                    assignCmd.execute(info.design());
                } finally {
                    setAsyncInProgress(false);
                }
            }
        });
        waitAsync();
    }

    protected AcmeElement findElement (AcmeDesign design, String elementName)
        throws ElementNotFoundException {

        final AcmeElement acmeE = (AcmeElement) design.lookupQualifiedName(elementName);
        if (acmeE == null) {
            pipe("X> Unable to find AcmeElement named " + elementName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_ELEMENT, elementName);
        }
        return acmeE;
    }

    protected AcmeSystem findSystem (AcmeDesign design, String sysName)
        throws ElementNotFoundException {

        final AcmeSystem sys = (AcmeSystem) design.lookupQualifiedName(sysName);
        if (sys == null) {
            pipe("X> Unable to find AcmeSystem named " + sysName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_SYS, sysName);
        }
        return sys;
    }

    protected AcmeRepresentation findRep (AcmeDesign design, String repName)
        throws ElementNotFoundException {

        final AcmeRepresentation rep = (AcmeRepresentation) design.lookupQualifiedName(repName);
        if (rep == null) {
            pipe("X> Unable to find AcmeRepresentation named " + repName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_REP, repName);
        }
        return rep;
    }

    protected AcmeComponent findComp (AcmeDesign design, String compName)
        throws ElementNotFoundException {

        final AcmeComponent comp = (AcmeComponent) design.lookupQualifiedName(compName);
        if (comp == null) {
            pipe("X> Unable to find AcmeComponent named " + compName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_COMP, compName);
        }
        return comp;
    }

    protected AcmeConnector findConn (AcmeDesign design, String connName)
        throws ElementNotFoundException {

        final AcmeConnector conn = (AcmeConnector) design.lookupQualifiedName(connName);
        if (conn == null) {
            pipe("X> Unable to find AcmeConnector named " + connName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_CONN, connName);
        }
        return conn;
    }

    protected AcmePort findPort (AcmeDesign design, String portName)
        throws ElementNotFoundException {
        
        final AcmePort port = (AcmePort) design.lookupQualifiedName(portName);
        if (port == null) {
            pipe("X> Unable to find AcmePort named " + portName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_PORT, portName);
        }
        return port;
    }

    protected AcmeRole findRole (AcmeDesign design, String roleName)
        throws ElementNotFoundException {

        final AcmeRole role = (AcmeRole) design.lookupQualifiedName(roleName);
        if (role == null) {
            pipe("X> Unable to find AcmeRole named " + roleName + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_ROLE, roleName);
        }
        return role;
    }

    protected AcmeProperty findProperty (AcmeDesign design, AcmeElement element, String propName)
        throws ElementNotFoundException {

        // first try finding within the element
        AcmeProperty prop = (AcmeProperty) element.lookupName(propName, true);
        if (prop == null) {  // then treat property as qualified name
            prop = (AcmeProperty) design.lookupQualifiedName(propName);
        }
        if (prop == null) {
            pipe("X> Unable to find AcmeProperty named " + propName
                + " in elemnt " + element.getQualifiedName() + " or design "
                + design.getQualifiedName() + "!");
            throw new ElementNotFoundException(ElementNotFoundException.NO_PROP, propName);
        }
        return prop;
    }

    protected AcmeFamily findFamily (AcmeDesign design, String fam) throws TypeNotFoundException {
        final AcmeFamily famT = (AcmeFamily) design.lookupName(fam, true);
        if (famT == null) {
            pipe("X> Unable to find AcmeFamily named " + fam + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_FAMILY, fam);
        }
        return famT;
    }

    protected AcmeElementType findElementType (AcmeSystem sys, String type)
        throws TypeNotFoundException {

        final AcmeElementType elementT = (AcmeElementType) sys.lookupName(type, true);
        if (elementT == null) {
            pipe("X> Unable to find AcmeElementType named " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_ELEMENT_TYPE, type);
        }
        return elementT;
    }

    protected AcmeComponentType findCompType (AcmeSystem sys, String type)
        throws TypeNotFoundException {

        final AcmeComponentType compT = (AcmeComponentType) sys.lookupName(type, true);
        if (compT == null) {
            pipe("X> Unable to find AcmeComponentType named " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_COMP_TYPE, type);
        }
        return compT;
    }

    protected AcmeConnectorType findConnType (AcmeSystem sys, String type)
        throws TypeNotFoundException {

        final AcmeConnectorType connT = (AcmeConnectorType) sys.lookupName(type, true);
        if (connT == null) {
            pipe("X> Unable to find AcmeConnectorType named " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_CONN_TYPE, type);
        }
        return connT;
    }

    protected AcmePortType findPortType (AcmeSystem sys, String type) throws TypeNotFoundException {
        final AcmePortType portT = (AcmePortType) sys.lookupName(type, true);
        if (portT == null) {
            pipe("X> Unable to find AcmePortType named " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_PORT_TYPE, type);
        }
        return portT;
    }

    protected AcmeRoleType findRoleType (AcmeSystem sys, String type) throws TypeNotFoundException {
        final AcmeRoleType roleT = (AcmeRoleType) sys.lookupName(type, true);
        if (roleT == null) {
            pipe("X> Unable to find AcmeRole named " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_ROLE_TYPE, type);
        }
        return roleT;
    }

    /**
     * Takes a property type name and property value, determines the actual
     * AcmePropertyType and the AcmePropertyValue, and returns those in an
     * anonymous AcmePropety.
     * @param sys  AcmeSystem in which to look up the property type
     * @param type  The type of the property
     * @param val  The value of the property
     * @return AcmeProperty  The anonymous AcmeProperty that contains the
     *   resolved AcmePropertyType and AcmePropertyValue
     * @throws NumberFormatException
     * @throws TypeNotFoundException
     */
    protected AcmeProperty determinePropVal (final AcmeSystem sys, String type, String val)
        throws NumberFormatException, TypeNotFoundException {
        // find the type referred to, and determine info value type
        AcmePropertyType propT = (AcmePropertyType) sys.lookupName(type, true);
        AcmePropertyValue propV = null;
        if (propT == null) {  // determine a basic type
            if (type.equalsIgnoreCase("int")) {
                propT = ConstantTypes.INT_TYPE;
                propV = new AcmeInt((IntType) propT, Integer.parseInt(val));
            } else if (type.equalsIgnoreCase("float")) {
                propT = ConstantTypes.FLOAT_TYPE;
                propV = new AcmeFloat((FloatType) propT, Float.parseFloat(val));
            } else if (type.equalsIgnoreCase("boolean")) {
                propT = ConstantTypes.BOOLEAN_TYPE;
                propV = new AcmeBoolean((BooleanType) propT, Boolean.getBoolean(val));
            } else if (type.equalsIgnoreCase("string")) {
                propT = ConstantTypes.STRING_TYPE;
                propV = new AcmeString((StringType) propT, val);
            } else if (type.equalsIgnoreCase("enum")) {
                propT = ConstantTypes.EMPTY_ENUM_TYPE;
                propV = new AcmeEnum((EnumType) propT, val);
            } else if (type.equalsIgnoreCase("set")) {
                propT = ConstantTypes.ANY_SET_TYPE;
            } else if (type.equalsIgnoreCase("sequence")) {
                propT = ConstantTypes.ANY_SEQUENCE_TYPE;
            } else if (type.equalsIgnoreCase("record")) {
                propT = ConstantTypes.ANY_RECORD_TYPE;
            } else if (type.equalsIgnoreCase("any")) {
                propT = ConstantTypes.EXPLICIT_ANY_TYPE;
                propV = new AcmeString(val);
            } else {
                propT = ConstantTypes.UNSPECIFIED_TYPE;
                propV = new AcmeString(val);
            }
        }
        if (propT == null) {  // shouldn't happen
            pipe("X> Unable to determine AcmePropertyType for " + type + "!");
            throw new TypeNotFoundException(TypeNotFoundException.NO_PROP_TYPE, type);
        }
        AcmeProperty prop = new AcmeProperty("_TEMP_", propT, propV);
        return prop;
    }

}
