package AcmeStudio.remote.jms;

import java.io.IOException;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
/**
 * Class: XAcmeChangeExporter <br>
 * Purpose: Takes an AcmeLib structure and converts it to the <br>
 * xAcme XML format <br>
 * Revision History: <br>
 * Date Person Change: <br>
 * 5/3/01 BRS Created initial version <br>
 * 
 * 
 * Known bugs/omissions: <br>- Should create a DOM structure and then serialize
 * it
 * 
 * @author Bradley Schmerl
 */
public class XAcmeChangeExporter {

	private class HashMap {

		// Hack - the MS JVM doesn't have the Hashmap
		private Vector	elements;

		public HashMap() {
			elements = new Vector ();
		}

		public Object get(Object key) {
			boolean found = false;
			int i = 0;
			Object element = null;
			while (i < elements.size () && !found) {
				Object comKey = elements.elementAt (i);
				if (comKey != null && comKey.equals (key)) {
					found = true;
					element = elements.elementAt (i + 1);
				}
				i += 2;
			}
			return element;
		}

		public void put(Object key, Object element) {
			elements.addElement (key);
			elements.addElement (element);
		}
	}
	private static final int	version	= 1;

	/**
	 * return the Acme identifier of a fully qualified unique xml identifier
	 * Assumed xml naming convention is: something.something.acmeName::number
	 * this function returns acmeName
	 * 
	 * @param complexName
	 *                    String
	 * @return the Acme name
	 */
	public static String decomposeName(String complexName) {
		// Returns the Acme name of a fully qualified Acme name
		String afterLastDot;
		if (complexName.lastIndexOf ('.') == -1) {
			afterLastDot = complexName;
		}
		else {
			afterLastDot = complexName.substring (complexName.lastIndexOf ('.') + 1, complexName.length ());
		}
		String upToColon;
		if (afterLastDot.indexOf (':') == -1) {
			upToColon = afterLastDot;
		}
		else {
			upToColon = afterLastDot.substring (0, afterLastDot.indexOf (':'));
		}
		return upToColon;
	}

	/**
	 * return the Acme identifier of a fully qualified unique xml identifier
	 * Assumed xml naming convention is: something.something.acmeName::number
	 * this function returns acmeName
	 * 
	 * @param complexName
	 *                    String
	 * @return the Acme name
	 */
	public static String getParentFullName(String complexName) {
		// Returns the Acme name of a fully qualified Acme name
		String upToLastDot, result;
		upToLastDot = complexName.substring (0, complexName.lastIndexOf ('.'));
		result = upToLastDot.substring (0, upToLastDot.length ());
		return result;
	}

	/**
	 * return the Acme identifier of a fully qualified unique xml identifier
	 * Assumed xml naming convention is: something.something.acmeName::number
	 * this function returns acmeName
	 * 
	 * @param complexName
	 *                    String
	 * @return the Acme name
	 */
	public static String getParentName(String complexName) {
		// Returns the Acme name of a fully qualified Acme name
		String upToLastDot, result;
		upToLastDot = complexName.substring (0, complexName.lastIndexOf ('.'));
		result = upToLastDot.substring (upToLastDot.lastIndexOf ('.') + 1, upToLastDot.length ());
		return result;
	}

	/**
	 * return the Acme identifier of a fully qualified unique xml identifier
	 * Assumed xml naming convention is: something.something.acmeName::number
	 * this function returns something.something.acmeName
	 * 
	 * @param complexName
	 *                    String
	 * @return the Acme name
	 */
	public static String removeColon(String complexName) {
		// Returns the Acme name of a fully qualified Acme name
		String afterLastDot = complexName;
		String upToColon;
		if (afterLastDot.indexOf (':') == -1) {
			upToColon = afterLastDot;
		}
		else {
			upToColon = afterLastDot.substring (0, afterLastDot.indexOf (':'));
		}
		return upToColon;
	}

	/**
	 * Strips the hash from the beginning of a name
	 * 
	 * @param nameWithHash
	 *                    String
	 * @return nameWithoutHash
	 */
	public static String stripHash(String nameWithHash) {
		String nameWithoutHash = nameWithHash;
		if (nameWithHash.indexOf ('#') != -1) {
			nameWithoutHash = nameWithHash.substring (nameWithHash.indexOf ('#') + 1, nameWithHash.length ());
		}
		return nameWithoutHash;
	}
	private final Document	_doc;
	private boolean			headerOutput			= false;
	private HashMap			idMap					= null;
	private boolean			includeVisProperties	= true;
	int						nextId					= 0;
	private HashMap			pidMap					= null;
	private HashMap			tidMap					= null;

	public XAcmeChangeExporter(Document doc) {
		_doc = doc;
		idMap = new HashMap ();
		pidMap = new HashMap ();
		tidMap = new HashMap ();
	}

	public void exportAttachedConnector(Element parentXML, String roleName, String portName, String tag) {
		Element xml = _doc.createElement (tag);
		xml.appendChild (exportRef (roleName, "roleName"));
		xml.appendChild (exportRef (portName, "portName"));
		parentXML.appendChild (xml);
	}

	public void exportAttachedGauge(Element parentXML, String name, String property, String mapping, String tag) {
		Element agXML = exportAttachedGauge (name, property, mapping, tag);
		parentXML.appendChild (agXML);
	}

	/**
	 * @param name
	 * @param property
	 * @param mapping
	 * @param tag
	 * @return
	 */
	private Element exportAttachedGauge(String name, String property, String mapping, String tag) {
		Element agXML = _doc.createElement (tag);
		Element nameXML = _doc.createElement ("name");
		nameXML.appendChild (_doc.createTextNode (name));
		agXML.appendChild (nameXML);
		agXML.appendChild (exportRef (property, XAcmeConstants.PROPERTY_TAG));
		if (mapping != null) {
			Element mapXML = _doc.createElement ("mapping");
			mapXML.appendChild (_doc.createTextNode (mapping));
			agXML.appendChild (mapXML);
		}
		return agXML;
	}

	/**
	 * @param attACME
	 * @return
	 */
	private Element exportAttachment(AcmeAttachment attACME) {
		Element linkXML = _doc.createElement (XAcmeConstants.LINKINSTANCE_TAG);
		linkXML.setAttribute ("id", generateID (null));
		linkXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		Element pointXML = _doc.createElement (XAcmeConstants.POINT_TAG);
		Element anchorXML = _doc.createElement (XAcmeConstants.ANCHORINTERFACE_TAG);
		anchorXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
		anchorXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + idMap.get (attACME.getPort ()));
		pointXML.appendChild (anchorXML);
		linkXML.appendChild (anchorXML);
		pointXML = _doc.createElement (XAcmeConstants.POINT_TAG);
		anchorXML = _doc.createElement (XAcmeConstants.ANCHORINTERFACE_TAG);
		anchorXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
		anchorXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + idMap.get (attACME.getRole ()));
		pointXML.appendChild (anchorXML);
		linkXML.appendChild (anchorXML);
		return linkXML;
	}

	public void exportAttachment(Element parentXML, AcmeAttachment attACME) {
		Element linkXML = exportAttachment (attACME);
		parentXML.appendChild (linkXML);
	}

	public void exportAttachment(Element parentXML, String port, String role, String tag) {
		Element attXML = exportAttachment (port, role, tag);
		parentXML.appendChild (attXML);
	}

	/**
	 * @param port
	 * @param role
	 * @param tag
	 * @return
	 */
	private Element exportAttachment(String port, String role, String tag) {
		Element attXML = _doc.createElement (tag);
		attXML.setAttribute ("id", generateID (null));
		attXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		Element anchorXML = _doc.createElement (XAcmeConstants.ANCHORINTERFACE_TAG);
		anchorXML.setAttribute ("id", port);
		attXML.appendChild (anchorXML);
		anchorXML = _doc.createElement (XAcmeConstants.ANCHORINTERFACE_TAG);
		anchorXML.setAttribute ("id", role);
		attXML.appendChild (anchorXML);
		return attXML;
	}

	public void exportBinding(Element parentXML, String inner, String outer, String tag) {
		exportAttachment (parentXML, inner, outer, tag);
	}

	public Element exportComponent(AcmeComponent componentACME) {
		return exportComponent (componentACME, XAcmeConstants.COMPONENTINSTANCE_TAG, true);
	}

	/**
	 * @param componentACME
	 * @param tag
	 * @param determineType
	 * @return
	 * @throws IOException
	 */
	private Element exportComponent(AcmeComponent componentACME, String tag, boolean determineType) {
		Element componentXML = _doc.createElement (tag);
		componentXML.setAttribute ("name", componentACME.getName ());
		componentXML.setAttribute ("id", generateID (componentACME));
		if (determineType && hasTypes (componentACME)) {
			componentXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "families:TypedComponentInstance");
		}
		else if (determineType
				&& (componentACME.getProperties ().getSize () > 0 || componentACME.getRepresentations ().getSize () > 0)) {
			componentXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "properties:ComponentInstance");
		}
		componentXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		AcmeEnumeration portsACME = componentACME.getPorts ();
		while (portsACME.moreItems ()) {
			AcmePort portACME = (AcmePort )portsACME.getNextItem ();
			exportPort (componentXML, portACME);
		}
		AcmeEnumeration representationsACME = componentACME.getRepresentations ();
		while (representationsACME.moreItems ()) {
			AcmeRepresentation representationACME = (AcmeRepresentation )representationsACME.getNextItem ();
			exportRepresentation (componentXML, representationACME);
		}
		AcmeEnumeration propertiesACME = componentACME.getProperties ();
		while (propertiesACME.moreItems ()) {
			AcmeProperty propertyACME = (AcmeProperty )propertiesACME.getNextItem ();
			exportProperty (componentXML, propertyACME);
		}
		if (hasTypes (componentACME)) {
			exportTypes (componentXML, componentACME, XAcmeConstants.TYPE_TAG);
			Element instanceOfXML = _doc.createElement (XAcmeConstants.INSTANCEOF_TAG);
			componentXML.appendChild (instanceOfXML);
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#"
					+ ((AcmeComponentType )componentACME.getInstantiatedTypes ().getNextItem ()).getTypeName ());
		}
		return componentXML;
	}

	public void exportComponent(Element parentXML, AcmeComponent componentACME) {
		Element compXML = exportComponent (componentACME);
		parentXML.appendChild (compXML);
	}

	public void exportComponent(Element parentXML, AcmeComponent componentACME, String tag, boolean determineType) {
		Element componentXML = exportComponent (componentACME, tag, determineType);
		parentXML.appendChild (componentXML);
	}

	/**
	 * @param componentTypeACME
	 * @return
	 * @throws IOException
	 */
	private Element exportComponentType(AcmeComponentType componentTypeACME) {
		Element componentTypeXML = _doc.createElement (XAcmeConstants.COMPONENTTYPE_TAG);
		componentTypeXML.setAttribute ("id", generateTypeID (componentTypeACME));
		exportSuperTypes (componentTypeXML, componentTypeACME, XAcmeConstants.SUPERTYPE_TAG);
		Element structureXML = exportComponent (componentTypeACME, XAcmeConstants.STRUCTURE_TAG, false);
		componentTypeXML.appendChild (structureXML);
		return componentTypeXML;
	}

	private void exportComponentType(Element parentXML, AcmeComponentType componentTypeACME) {
		Element componentTypeXML = exportComponentType (componentTypeACME);
		parentXML.appendChild (componentTypeXML);
	}

	public Element exportConnector(AcmeConnector connectorACME) {
		return exportConnector (connectorACME, XAcmeConstants.CONNECTORINSTANCE_TAG, true);
	}

	/**
	 * @param connectorACME
	 * @param tag
	 * @param determineType
	 * @return
	 * @throws IOException
	 */
	private Element exportConnector(AcmeConnector connectorACME, String tag, boolean determineType) {
		Element connectorXML = _doc.createElement (tag);
		connectorXML.setAttribute ("name", connectorACME.getName ());
		connectorXML.setAttribute ("id", generateID (connectorACME));
		if (determineType && hasTypes (connectorACME)) {
			connectorXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "families:TypedConnectorInstance");
		}
		else if (determineType
				&& (connectorACME.getProperties ().getSize () > 0 || connectorACME.getRepresentations ().getSize () > 0)) {
			connectorXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "properties:ConnectorInstance");
		}
		connectorXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		AcmeEnumeration rolesACME = connectorACME.getRoles ();
		while (rolesACME.moreItems ()) {
			AcmeRole roleACME = (AcmeRole )rolesACME.getNextItem ();
			exportRole (connectorXML, roleACME);
		}
		AcmeEnumeration representationsACME = connectorACME.getRepresentations ();
		while (representationsACME.moreItems ()) {
			AcmeRepresentation representationACME = (AcmeRepresentation )representationsACME.getNextItem ();
			exportRepresentation (connectorXML, representationACME);
		}
		AcmeEnumeration propertiesACME = connectorACME.getProperties ();
		while (propertiesACME.moreItems ()) {
			AcmeProperty propertyACME = (AcmeProperty )propertiesACME.getNextItem ();
			exportProperty (connectorXML, propertyACME);
		}
		if (hasTypes (connectorACME)) {
			exportTypes (connectorXML, connectorACME, XAcmeConstants.TYPE_TAG);
			Element instanceOfXML = _doc.createElement (XAcmeConstants.INSTANCEOF_TAG);
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#"
					+ ((AcmeConnectorType )connectorACME.getInstantiatedTypes ().getNextItem ()).getTypeName ());
			connectorXML.appendChild (instanceOfXML);
		}
		return connectorXML;
	}

	public void exportConnector(Element parentXML, AcmeConnector connectorACME) {
		Element connectorXML = exportConnector (connectorACME, XAcmeConstants.CONNECTORINSTANCE_TAG, true);
		parentXML.appendChild (connectorXML);
	}

	public void exportConnector(Element parentXML, AcmeConnector connectorACME, String tag, boolean determineType) {
		Element connectorXML = exportConnector (connectorACME, tag, determineType);
		parentXML.appendChild (connectorXML);
	}

	/**
	 * @param connectorTypeACME
	 * @return
	 * @throws IOException
	 */
	private Element exportConnectorType(AcmeConnectorType connectorTypeACME) {
		Element connectorTypeXML = _doc.createElement (XAcmeConstants.CONNECTORTYPE_TAG);
		connectorTypeXML.setAttribute ("id", generateTypeID (connectorTypeACME));
		exportSuperTypes (connectorTypeXML, connectorTypeACME, XAcmeConstants.SUPERTYPE_TAG);
		Element structureXML = exportConnector (connectorTypeACME, XAcmeConstants.STRUCTURE_TAG, false);
		connectorTypeXML.appendChild (structureXML);
		return connectorTypeXML;
	}

	private void exportConnectorType(Element parentXML, AcmeConnectorType connectorTypeACME) {
		Element connectorTypeXML = exportConnectorType (connectorTypeACME);
		parentXML.appendChild (connectorTypeXML);
	}

	/**
	 * @param familyACME
	 * @return
	 * @throws IOException
	 */
	private Element exportFamily(AcmeFamily familyACME) {
		Element familyXML = _doc.createElement (XAcmeConstants.FAMILY_TAG);
		familyXML.setAttribute ("id", generateTypeID (familyACME));
		exportSuperTypes (familyXML, familyACME, XAcmeConstants.SUPERTYPE_TAG);
		exportTypespace (familyXML, familyACME.getTypes ());
		Element structure = exportSystem (familyACME, XAcmeConstants.STRUCTURE_TAG, false);
		familyXML.appendChild (structure);
		return familyXML;
	}

	private void exportFamily(Element parentXML, AcmeFamily familyACME) {
		Element familyXML = exportFamily (familyACME);
		parentXML.appendChild (familyXML);
	}

	public void exportGauge(Element parentXML, String gaugeName, AcmeProperty property, String tag) {
		Element xml = exportGauge (gaugeName, property, tag);
		parentXML.appendChild (xml);
	}

	public void exportGauge(Element parentXML, String gaugeName, AcmeProperty property, String mapping, AcmeSystem system,
			String tag) {
		Element xml = exportGauge (gaugeName, property, mapping, system, tag);
		parentXML.appendChild (xml);
	}

	public void exportGauge(Element parentXML, String gaugeName, String gaugeType, AcmeSystem system, String tag) {
		Element xml = exportGauge (gaugeName, gaugeType, system, tag);
		parentXML.appendChild (xml);
	}

	/**
	 * @param gaugeName
	 * @param property
	 * @param tag
	 * @return
	 */
	private Element exportGauge(String gaugeName, AcmeProperty property, String tag) {
		Element xml = _doc.createElement (tag);
		xml.setAttribute ("name", gaugeName);
		xml.setAttribute (XAcmeConstants.PROPERTY_TAG, property.getQualifiedName ());
		xml.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		return xml;
	}

	/**
	 * @param gaugeName
	 * @param property
	 * @param mapping
	 * @param system
	 * @param tag
	 * @return
	 */
	private Element exportGauge(String gaugeName, AcmeProperty property, String mapping, AcmeSystem system, String tag) {
		Element xml = _doc.createElement (tag);
		xml.setAttribute ("name", gaugeName);
		xml.setAttribute ("id", system.getName () + "." + gaugeName);
		xml.setAttribute (XAcmeConstants.PROPERTY_TAG, property.getName ());
		xml.setAttribute ("mapping", mapping);
		xml.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		return xml;
	}

	/**
	 * @param gaugeName
	 * @param gaugeType
	 * @param system
	 * @param tag
	 * @return
	 */
	private Element exportGauge(String gaugeName, String gaugeType, AcmeSystem system, String tag) {
		Element xml = _doc.createElement (tag);
		xml.setAttribute ("name", gaugeName);
		xml.setAttribute ("id", system.getName () + "." + gaugeName);
		xml.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		Element typeXML = _doc.createElement (XAcmeConstants.TYPE_TAG);
		xml.appendChild (typeXML);
		typeXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
		typeXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + gaugeType);
		return xml;
	}

	public void exportNewGauge(Element parentXML, String name, String type, String tag) {
		Element xml = exportNewGauge (name, type, tag);
		parentXML.appendChild (xml);
	}

	/**
	 * @param name
	 * @param type
	 * @param tag
	 * @return
	 */
	private Element exportNewGauge(String name, String type, String tag) {
		Element xml = _doc.createElement (tag);
		Element refXML = exportRef (name, "name");
		xml.appendChild (xml);
		Element typeXML = _doc.createElement (XAcmeConstants.TYPE_TAG);
		typeXML.appendChild (_doc.createTextNode (type));
		xml.appendChild (typeXML);
		return xml;
	}

	public Element exportPort(AcmePort portACME) {
		return exportPort (portACME, XAcmeConstants.INTERFACEINSTANCE_TAG, true);
	}

	/**
	 * @param portACME
	 * @param tag
	 * @param determineType
	 * @return
	 * @throws IOException
	 */
	private Element exportPort(AcmePort portACME, String tag, boolean determineType) {
		Element portXML = _doc.createElement (tag);
		portXML.setAttribute ("name", portACME.getName ());
		portXML.setAttribute ("id", generateID (portACME));
		if (determineType && hasTypes (portACME)) {
			portXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "families:TypedInterfaceInstance");
		}
		else if (determineType && portACME.getProperties ().getSize () > 0) {
			portXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "properties:InterfaceInstance");
		}
		portXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		AcmeEnumeration propertiesACME = portACME.getProperties ();
		while (propertiesACME.moreItems ()) {
			AcmeProperty propertyACME = (AcmeProperty )propertiesACME.getNextItem ();
			exportProperty (portXML, propertyACME);
		}
		if (hasTypes (portACME)) {
			exportTypes (portXML, portACME, XAcmeConstants.TYPE_TAG);
			Element instanceOfXML = _doc.createElement (XAcmeConstants.INSTANCEOF_TAG);
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#"
					+ ((AcmePortType )portACME.getInstantiatedTypes ().getNextItem ()).getTypeName ());
		}
		return portXML;
	}

	public void exportPort(Element parentXML, AcmePort portACME) {
		Element portXML = exportPort (portACME);
		parentXML.appendChild (portXML);
	}

	public void exportPort(Element parentXML, AcmePort portACME, String tag, boolean determineType) {
		Element portXML = exportPort (portACME, tag, determineType);
		parentXML.appendChild (portXML);
	}

	/**
	 * @param portTypeACME
	 * @return
	 * @throws IOException
	 */
	private Element exportPortType(AcmePortType portTypeACME) {
		Element portTypeXML = _doc.createElement (XAcmeConstants.INTERFACETYPE_TAG);
		portTypeXML.setAttribute ("id", generateTypeID (portTypeACME));
		exportSuperTypes (portTypeXML, portTypeACME, XAcmeConstants.SUPERTYPE_TAG);
		Element structureXML = exportPort (portTypeACME, XAcmeConstants.STRUCTURE_TAG, false);
		portTypeXML.appendChild (structureXML);
		return portTypeXML;
	}

	private void exportPortType(Element parentXML, AcmePortType portTypeACME) {
		Element portTypeXML = exportPortType (portTypeACME);
		parentXML.appendChild (portTypeXML);
	}

	public void exportProperty(Element parentXML, AcmeProperty propertyACME) {
		exportProperty (parentXML, propertyACME, XAcmeConstants.PROPERTY_TAG, null);
	}

	public void exportProperty(Element parentXML, AcmeProperty propertyACME, String tag, String hostName) {
		String id = generatePropertyID (propertyACME);
		if (id.indexOf (".") == -1 && hostName != null) {
			id = hostName + "." + id;
		}
		if (includeProperty (propertyACME)) {
			Element propertyXML = _doc.createElement (tag);
			propertyXML.setAttribute ("name", propertyACME.getName ());
			propertyXML.setAttribute ("id", id);
			AcmePropertyValue valueACME = propertyACME.getValue ();
			if (valueACME != null) {
				exportPropertyValue (propertyXML, valueACME);
			}
			exportPropertyType (propertyXML, propertyACME.getType (), true);
			parentXML.appendChild (propertyXML);
		}
	}

	private void exportPropertyType(Element parentXML, AcmePropertyType propertyTypeACME, boolean displayTag) {
		Element realParentXML = parentXML;
		if (displayTag) {
			Element propertyTypeXML = _doc.createElement (XAcmeConstants.PROPERTYTYPE_TAG);
			propertyTypeXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "acmeProperties:PropertyType");
			parentXML.appendChild (propertyTypeXML);
			realParentXML = propertyTypeXML;
		}
		if (propertyTypeACME instanceof acmeLib.properties.RecordType) {
			Element propertyRecordTypeXML = _doc.createElement (XAcmeConstants.PROPERTYRECORDTYPE_TAG);
			acmeLib.properties.RecordType recordACME = (acmeLib.properties.RecordType )propertyTypeACME;
			AcmeEnumeration fieldsACME = recordACME.getFields ();
			while (fieldsACME.moreItems ()) {
				AcmeRecordField fieldACME = (AcmeRecordField )fieldsACME.getNextItem ();
				Element fieldDeclarationXML = _doc.createElement (XAcmeConstants.PROPERTYFIELDDECLARATION_TAG);
				fieldDeclarationXML.setAttribute ("name", fieldACME.getName ());
				AcmePropertyType fieldTypeACME = fieldACME.getType ();
				exportPropertyType (fieldDeclarationXML, fieldTypeACME, true);
				propertyRecordTypeXML.appendChild (fieldDeclarationXML);
			}
			realParentXML.appendChild (propertyRecordTypeXML);
		}
		else if (propertyTypeACME instanceof acmeLib.properties.SetType) {
			Element propertySetTypeXML = _doc.createElement (XAcmeConstants.PROPERTYSETTYPE_TAG);
			acmeLib.properties.SetType setTypeACME = (acmeLib.properties.SetType )propertyTypeACME;
			AcmePropertyType innerTypeACME = (AcmePropertyType )setTypeACME.getSetType ();
			exportPropertyType (propertySetTypeXML, innerTypeACME, true);
			realParentXML.appendChild (propertySetTypeXML);
		}
		else if (propertyTypeACME instanceof acmeLib.properties.SequenceType) {
			Element propertySequenceTypeXML = _doc.createElement (XAcmeConstants.PROPERTYSEQUENCETYPE_TAG);
			acmeLib.properties.SequenceType seqTypeACME = (acmeLib.properties.SequenceType )propertyTypeACME;
			AcmePropertyType innerTypeACME = seqTypeACME.getSequenceType ();
			exportPropertyType (propertySequenceTypeXML, innerTypeACME, true);
			realParentXML.appendChild (propertySequenceTypeXML);
		}
		else if (propertyTypeACME instanceof EnumType) {
			Element propertyEnumTypeXML = _doc.createElement (XAcmeConstants.PROPERTYENUMTYE_TAG);
			EnumType enumTypeACME = (EnumType )propertyTypeACME;
			AcmeEnumeration valuesACME = enumTypeACME.getValues ();
			while (valuesACME.moreItems ()) {
				EnumValue valueACME = (EnumValue )valuesACME.getNextItem ();
				Element valueXML = _doc.createElement ("value");
				Text value = _doc.createTextNode (valueACME.value);
				valueXML.appendChild (value);
				propertyEnumTypeXML.appendChild (valueXML);
			}
			realParentXML.appendChild (propertyEnumTypeXML);
		}
		else if (propertyTypeACME instanceof StringType) {
			Element primTypeXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVETYPE_TAG);
			primTypeXML.setAttribute (XAcmeConstants.TYPE_TAG, "string");
			realParentXML.appendChild (primTypeXML);
		}
		else if (propertyTypeACME instanceof IntType) {
			Element primTypeXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVETYPE_TAG);
			primTypeXML.setAttribute (XAcmeConstants.TYPE_TAG, "int");
			realParentXML.appendChild (primTypeXML);
		}
		else if (propertyTypeACME instanceof FloatType) {
			Element primTypeXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVETYPE_TAG);
			primTypeXML.setAttribute (XAcmeConstants.TYPE_TAG, "float");
			realParentXML.appendChild (primTypeXML);
		}
		else if (propertyTypeACME instanceof BooleanType) {
			Element primTypeXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVETYPE_TAG);
			primTypeXML.setAttribute (XAcmeConstants.TYPE_TAG, "boolean");
			realParentXML.appendChild (primTypeXML);
		}
	}

	private void exportPropertyTypeDecl(Element parentXML, AcmePropertyType propertyTypeACME) {
		Element propertyTypeXML = _doc.createElement (XAcmeConstants.PROPERTYTYPE_TAG);
		propertyTypeXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "acmeProperties:PropertyType");
		propertyTypeXML.setAttribute ("name", generateTypeID (propertyTypeACME));
		exportPropertyType (propertyTypeXML, propertyTypeACME, false);
		parentXML.appendChild (propertyTypeXML);
	}

	/**
	 * @param valueACME
	 * @return
	 */
	private Element exportPropertyValue(AcmePropertyValue valueACME) {
		Element valueXML = _doc.createElement ("value");
		valueXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "acmeProperties:PropertyValue");
		if (valueACME instanceof AcmeString) {
			AcmeString stringValACME = (AcmeString )valueACME;
			Element ppvXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVEVALUE_TAG);
			Element valXML = _doc.createElement (XAcmeConstants.STRINGVALUE_TAG);
			ppvXML.appendChild (valXML);
			valXML.appendChild (_doc.createTextNode (stringValACME.getValue ()));
			valueXML.appendChild (ppvXML);
		}
		else if (valueACME instanceof AcmeInt) {
			AcmeInt intValACME = (AcmeInt )valueACME;
			Element ppvXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVEVALUE_TAG);
			Element valXML = _doc.createElement (XAcmeConstants.INTVALUE_TAG);
			ppvXML.appendChild (valXML);
			valXML.appendChild (_doc.createTextNode (String.valueOf (intValACME.getValue ())));
			valueXML.appendChild (ppvXML);
		}
		else if (valueACME instanceof AcmeFloat) {
			AcmeFloat floatValACME = (AcmeFloat )valueACME;
			Element ppvXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVEVALUE_TAG);
			Element valXML = _doc.createElement (XAcmeConstants.FLOATVALUE_TAG);
			ppvXML.appendChild (valXML);
			valXML.appendChild (_doc.createTextNode (String.valueOf (floatValACME.getValue ())));
			valueXML.appendChild (ppvXML);
		}
		else if (valueACME instanceof AcmeBoolean) {
			AcmeBoolean booleanValACME = (AcmeBoolean )valueACME;
			Element ppvXML = _doc.createElement (XAcmeConstants.PROPERTYPRIMITIVEVALUE_TAG);
			Element valXML = _doc.createElement (XAcmeConstants.BOOLEANVALUE_TAG);
			ppvXML.appendChild (valXML);
			valXML.appendChild (_doc.createTextNode (String.valueOf (booleanValACME.getValue ())));
			valueXML.appendChild (ppvXML);
		}
		else if (valueACME instanceof AcmeSet) {
			AcmeSet setValACME = (AcmeSet )valueACME;
			Element setValXML = _doc.createElement (XAcmeConstants.PROPERTYSETVALUE_TAG);
			AcmeEnumeration valuesACME = setValACME.getValues ();
			while (valuesACME.moreItems ()) {
				AcmePropertyValue vACME = (AcmePropertyValue )valuesACME.getNextItem ();
				exportPropertyValue (setValXML, vACME);
			}
			valueXML.appendChild (setValXML);
		}
		else if (valueACME instanceof AcmeSequence) {
			AcmeSequence sequenceValACME = (AcmeSequence )valueACME;
			Element seqValXML = _doc.createElement (XAcmeConstants.PROPERTYSEQUENCEVALUE_TAG);
			AcmeEnumeration valuesACME = sequenceValACME.getValues ();
			while (valuesACME.moreItems ()) {
				AcmePropertyValue vACME = (AcmePropertyValue )valuesACME.getNextItem ();
				exportPropertyValue (seqValXML, vACME);
			}
			valueXML.appendChild (seqValXML);
		}
		else if (valueACME instanceof AcmeEnum) {
		}
		else if (valueACME instanceof AcmeRecord) {
			AcmeRecord recACME = (AcmeRecord )valueACME;
			Element recordValXML = _doc.createElement (XAcmeConstants.PROPERTYRECORDVALUE_TAG);
			AcmeEnumeration fieldsACME = recACME.getFields ();
			while (fieldsACME.moreItems ()) {
				AcmeRecordField fieldACME = (AcmeRecordField )fieldsACME.getNextItem ();
				Element fieldXML = _doc.createElement (XAcmeConstants.PROPERTYFIELDVALUE_TAG);
				fieldXML.setAttribute ("fieldName", fieldACME.getName ());
				exportPropertyValue (fieldXML, fieldACME.getValue ());
				recordValXML.appendChild (fieldXML);
			}
			valueXML.appendChild (recordValXML);
		}
		return valueXML;
	}

	public void exportPropertyValue(Element parentXML, AcmePropertyValue valueACME) {
		Element valueXML = exportPropertyValue (valueACME);
		parentXML.appendChild (valueXML);
	}

	public void exportRef(Element xml, String name, String tag) {
		Element refXML = exportRef (name, tag);
		xml.appendChild (refXML);
	}

	/**
	 * @param name
	 * @param tag
	 * @return
	 */
	private Element exportRef(String name, String tag) {
		Element refXML = _doc.createElement (tag);
		Element instanceOfXML = _doc.createElement (XAcmeConstants.INSTANCEOF_TAG);
		refXML.appendChild (instanceOfXML);
		instanceOfXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
		instanceOfXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + name);
		return refXML;
	}

	/**
	 * @param repACME
	 * @return
	 */
	private Element exportRepresentation(AcmeRepresentation repACME) {
		Element repXML = _doc.createElement (XAcmeConstants.SUBARCHITECTURE_TAG);
		exportSystem (repXML, repACME.getSystem (), XAcmeConstants.ARCHINSTANCE_TAG, true);
		AcmeEnumeration bindingsACME = ((AcmeBindingsMap )repACME.getAbstractionMap ()).getBindings ();
		while (bindingsACME.moreItems ()) {
			AcmeBinding bindingACME = (AcmeBinding )bindingsACME.getNextItem ();
			AcmeElement innerACME = bindingACME.getInner ();
			AcmeElement outerACME = bindingACME.getOuter ();
			Element iimXML = _doc.createElement (XAcmeConstants.INTERFACEINSTANCEMAPPING_TAG);
			Element oiiXML = _doc.createElement (XAcmeConstants.OUTERINTERFACEINSTANCE_TAG);
			oiiXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			oiiXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + idMap.get (outerACME));
			repXML.appendChild (oiiXML);
			Element iiiXML = _doc.createElement (XAcmeConstants.INNERINTERFACEINSTANCE_TAG);
			iiiXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			iiiXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + idMap.get (innerACME));
			repXML.appendChild (iiiXML);
		}
		return repXML;
	}

	private void exportRepresentation(Element parentXML, AcmeRepresentation repACME) {
		Element repXML = exportRepresentation (repACME);
		parentXML.appendChild (repXML);
	}

	public Element exportRole(AcmeRole roleACME) {
		return exportRole (roleACME, XAcmeConstants.INTERFACEINSTANCE_TAG, true);
	}

	/**
	 * @param roleACME
	 * @param tag
	 * @param determineType
	 * @return
	 */
	private Element exportRole(AcmeRole roleACME, String tag, boolean determineType) {
		Element roleXML = _doc.createElement (tag);
		roleXML.setAttribute ("name", roleACME.getName ());
		roleXML.setAttribute ("id", generateID (roleACME));
		if (determineType && hasTypes (roleACME)) {
			roleXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "families:TypedInterfaceInstance");
		}
		else if (determineType && roleACME.getProperties ().getSize () > 0) {
			roleXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "properties:InterfaceInstance");
		}
		roleXML.appendChild (_doc.createElement (XAcmeConstants.DESCRIPTION_TAG));
		AcmeEnumeration propertiesACME = roleACME.getProperties ();
		while (propertiesACME.moreItems ()) {
			AcmeProperty propertyACME = (AcmeProperty )propertiesACME.getNextItem ();
			exportProperty (roleXML, propertyACME);
		}
		if (hasTypes (roleACME)) {
			exportTypes (roleXML, roleACME, XAcmeConstants.TYPE_TAG);
			Element instanceOfXML = _doc.createElement (XAcmeConstants.INSTANCEOF_TAG);
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
			instanceOfXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#"
					+ ((AcmeRoleType )roleACME.getInstantiatedTypes ().getNextItem ()).getTypeName ());
		}
		return roleXML;
	}

	public void exportRole(Element parentXML, AcmeRole roleACME) {
		Element roleXML = exportRole (roleACME);
		parentXML.appendChild (roleXML);
	}

	public void exportRole(Element parentXML, AcmeRole roleACME, String tag, boolean determineType) {
		Element roleXML = exportRole (roleACME, tag, determineType);
		parentXML.appendChild (roleXML);
	}

	/**
	 * @param roleTypeACME
	 * @return
	 * @throws IOException
	 */
	private Element exportRoleType(AcmeRoleType roleTypeACME) {
		Element roleTypeXML = _doc.createElement (XAcmeConstants.INTERFACETYPE_TAG);
		roleTypeXML.setAttribute ("id", generateTypeID (roleTypeACME));
		exportSuperTypes (roleTypeXML, roleTypeACME, XAcmeConstants.SUPERTYPE_TAG);
		Element structureXML = exportRole (roleTypeACME, XAcmeConstants.STRUCTURE_TAG, false);
		roleTypeXML.appendChild (structureXML);
		return roleTypeXML;
	}

	private void exportRoleType(Element parentXML, AcmeRoleType roleTypeACME) {
		Element roleTypeXML = exportRoleType (roleTypeACME);
		parentXML.appendChild (roleTypeXML);
	}

	private void exportSuperTypes(Element parentXML, AcmeElementType elementTypeACME, String tag) {
		int nTypes = elementTypeACME.getSuperTypes ().getSize ();
		if (nTypes != 0) {
			AcmeEnumeration supersACME = elementTypeACME.getSuperTypes ().enumerate (
					new acmeLib.core.filters.NoBaseClassesEnumFilter ());
			while (supersACME.moreItems ()) {
				AcmeElementType superACME = (AcmeElementType )supersACME.getNextItem ();
				Element superXML = _doc.createElement (tag);
				superXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
				superXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + tidMap.get (superACME));
				parentXML.appendChild (superXML);
			}
		}
	}

	/**
	 * @param systemACME
	 * @param tag
	 * @param doSupers
	 * @return
	 */
	private Element exportSystem(AcmeSystem systemACME, String tag, boolean doSupers) {
		Element systemXML = _doc.createElement (tag);
		systemXML.setAttribute ("id", (tag.equals (XAcmeConstants.FAMILY_TAG)
				? generateTypeID ((AcmeFamily )systemACME)
				: generateID (systemACME)));
		if (doSupers && systemACME.getDeclaredTypes ().getSize () > 0) {
			systemXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "families:TypedArchInstance");
		}
		else if (!tag.equals (XAcmeConstants.FAMILY_TAG) && systemACME.getProperties ().getSize () > 0) {
			systemXML.setAttribute (XAcmeConstants.XSI_TYPE_ATTR, "properties:ArchInstance");
		}
		AcmeEnumeration componentsACME = systemACME.getComponents ();
		while (componentsACME.moreItems ()) {
			exportComponent (systemXML, (AcmeComponent )componentsACME.getNextItem ());
		}
		AcmeEnumeration connectorsACME = systemACME.getConnectors ();
		while (connectorsACME.moreItems ()) {
			exportConnector (systemXML, (AcmeConnector )connectorsACME.getNextItem ());
		}
		AcmeEnumeration attachmentsACME = systemACME.getAttachments ();
		int numAttachments = 0;
		while (attachmentsACME.moreItems ()) {
			exportAttachment (systemXML, (AcmeAttachment )attachmentsACME.getNextItem ());
		}
		if (doSupers) {
			this.exportTypes (systemXML, systemACME, XAcmeConstants.TYPE_TAG);
		}
		return systemXML;
	}

	private void exportSystem(Element parentXML, AcmeSystem systemACME, String tag, boolean doSupers) {
		Element systemXML = exportSystem (systemACME, tag, doSupers);
		parentXML.appendChild (systemXML);
	}

	private void exportTypes(Element parentXML, AcmeElement elementACME, String tag) {
		int nTypes = elementACME.getDeclaredTypes ().getSize ();
		if (nTypes != 0) {
			AcmeEnumeration supersACME = elementACME.getDeclaredTypes ();
			while (supersACME.moreItems ()) {
				AcmeElementType superACME = (AcmeElementType )supersACME.getNextItem ();
				Element superXML = _doc.createElement (tag);
				superXML.setAttribute (XAcmeConstants.XLINK_TYPE_ATTR, "simple");
				superXML.setAttribute (XAcmeConstants.XLINK_HREF_ATTR, "#" + superACME.getTypeName ());
				parentXML.appendChild (superXML);
			}
		}
	}

	//  public void exportDesign (AcmeDesign designACME)
	//    throws java.io.IOException
	//  {
	//  		_doc =
	//  	
	//    out.writeln ("<?xml version=\"1.0\"?>");
	//    out.writeln ("<!-- Automatically generated by XAcmeChangeExporter " +
	// version +
	//		 " using xAcme version 1.0 -->");
	//    out.write ("<!-- Generated ");
	//    GregorianCalendar cal = new GregorianCalendar (TimeZone.getDefault ());
	//    out.writeln (cal.get (Calendar.MONTH) + "/" +
	//		 cal.get (Calendar.DAY_OF_MONTH) + "/" +
	//		 cal.get (Calendar.YEAR) + " -->");
	//    
	//    out.write ("<typedDesign ");
	//    outputHeader ();
	//    out.writeln (">");
	//    
	//    exportTypespace (designACME.getTypes ());
	//    
	//    AcmeEnumeration systemsACME = designACME.getSystems ();
	//    while (systemsACME.moreItems ()) {
	//      AcmeSystem systemACME = (AcmeSystem )systemsACME.getNextItem ();
	//      exportSystem (systemACME, "archInstance", true, true);
	//    }
	//    
	//    out.writeln ("</typedDesign>");
	//    out.close ();
	//  }
	private void exportTypespace(Element parentXML, AcmeEnumeration typespace) {
		typespace.sortByCreationOrder ();
		while (typespace.moreItems ()) {
			AcmeType typeACME = (AcmeType )typespace.getNextItem ();
			if (typeACME instanceof AcmePropertyType) {
				exportPropertyTypeDecl (parentXML, (AcmePropertyType )typeACME);
			}
		}
		typespace.reset ();
		while (typespace.moreItems ()) {
			AcmeType typeACME = (AcmeType )typespace.getNextItem ();
			if (typeACME instanceof AcmeComponentType) {
				exportComponentType (parentXML, (AcmeComponentType )typeACME);
			}
		}
		typespace.reset ();
		while (typespace.moreItems ()) {
			AcmeType typeACME = (AcmeType )typespace.getNextItem ();
			if (typeACME instanceof AcmeConnectorType) {
				exportConnectorType (parentXML, (AcmeConnectorType )typeACME);
			}
		}
		typespace.reset ();
		while (typespace.moreItems ()) {
			AcmeType typeACME = (AcmeType )typespace.getNextItem ();
			if (typeACME instanceof AcmePortType) {
				exportPortType (parentXML, (AcmePortType )typeACME);
			}
			else if (typeACME instanceof AcmeRoleType) {
				exportRoleType (parentXML, (AcmeRoleType )typeACME);
			}
		}
		typespace.reset ();
		while (typespace.moreItems ()) {
			AcmeType typeACME = (AcmeType )typespace.getNextItem ();
			if (typeACME instanceof AcmeFamily) {
				exportFamily (parentXML, (AcmeFamily )typeACME);
			}
		}
	}

	private String generateFlatName(AcmeElement elementACME) {
		return elementACME.getQualifiedName ();
	}

	private String generateID(AcmeElement elementACME) {
		String id = "";
		if (elementACME != null) id = generateFlatName (elementACME);
		id += "::" + nextId++;
		idMap.put (elementACME, id);
		return id;
	}

	private String generatePropertyID(AcmeProperty propertyACME) {
		String id = propertyACME.getQualifiedName () + "::" + nextId++;
		pidMap.put (propertyACME, id);
		return id;
	}

	private String generateTypeID(AcmeType typeACME) {
		String id = typeACME.getTypeName () + "::" + nextId++;
		tidMap.put (typeACME, id);
		return id;
	}

	private boolean hasTypes(AcmeElement elementACME) {
		return elementACME.getDeclaredTypes ().getSize () > 0 || elementACME.getInstantiatedTypes ().getSize () > 0;
	}

	private boolean includeProperty(AcmeProperty property) {
		boolean inc = true;
		return inc;
		/*
		 * if (!includeVisProperties) { String name = property.getName (); if
		 * (name.length () >= 4 && name.substring (0,4).equals ("vis-")){ inc =
		 * false; } } return inc;
		 */
	}

	public static void outputHeader(Element xml)  {
		xml.setAttribute ("xmlns", "http://www.ics.uci.edu/pub/arch/xArch/instance.xsd");
		xml.setAttribute ("xmlns:xsi", "http://www.w3.org/1999/XMLSchema/instance");
		xml.setAttribute ("xmlns:instance", "http://www.ics.uci.edu/pub/arch/xArch/instance.xsd");
		xml.setAttribute ("xmlns:properties", "http://www.cs.cmu.edu/~acme/pub/xAcme/properties.xsd");
		xml.setAttribute ("xmlns:acmeProperties", "http://www.cs.cmu.edu/~acme/pub/xAcme/acmeProperties.xsd");
		xml.setAttribute ("xmlns:families", "http://www.cs.cmu.edu/~acme/pub/xAcme/families.xsd");
		xml.setAttribute ("xmlns:constraints", "http://www.cs.cmu.edu/~acme/pub/xAcme/constraints.xsd");
		xml.setAttribute ("xmlns:xlink", "http://www.w3.org/1999/xlink");
		xml.setAttribute ("xsi:schemaLocation", "http://www.w3.org/1999/xlink");
	}
}
