/*
 * Created on Jul 24, 2003
 */
package AcmeStudio.remote.test.statement;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.Remote;
import java.util.*;


/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * Base class for statements that can be run in RemoteControlTest from prompt. 
 */
public class Statement {

    private static final String REMOTE_VAR = "rmdz";

    protected static Map _varMap = new HashMap();

    protected String _retVar = null;

    protected String _server = null;

    protected String _cmd = null;

    protected Object[] _params = null;
    protected Class[] _paramTypes = null;

    public Statement (String cmd) {
        _cmd = cmd;
    }

    public static Statement parse (String line) {
        Statement stmt = null;
        String var = null;
        int idx = line.indexOf("=");
        if (idx > 0) {  // var must at least have a name
            var = line.substring(0, idx).trim();
            line = line.substring(idx+1).trim();
        } else {
            line = line.trim();
        }

        // find the first space, between command and rest of params, if any 
        int firstSpace = line.indexOf(" ");
        String cmd = null;
        String paramStr = null;
        if (firstSpace > 0) {
            cmd = line.substring(0, firstSpace).trim();
            paramStr = line.substring(firstSpace+1);
        } else {
            cmd = line;
        }

        // look for a dot before command, for statements that access a remote design
        int dotIdx = cmd.indexOf(".");
        String server = null;
        if (dotIdx > 0) {
            server = line.substring(0, dotIdx).trim();
            cmd = cmd.substring(dotIdx+1).trim();
        }

        // construct a statement object
        if (server == null && StatementFactory.isLocalStatement(cmd)) {
            stmt = StatementFactory.newStatement(cmd);
        }
        if (stmt == null) {
            stmt = new Statement(cmd);
        }
        stmt.setReturnVar(var);
        stmt.setServer(server);

        if (paramStr != null) {
            // extract the space-delimited parameters
            // scan through the string a character at a time,
            //   keeping track of double quotes for strings, which may
            //   contain space, and treat strings as one unit
            List paramList = new ArrayList();
            // indicate beginning index of parameter
            int i, paramStart;
            // indicates that end of string quote not yet encountered
            boolean openQuote = false;
            for (i=0, paramStart=0; i < paramStr.length(); i++) {
                char c = paramStr.charAt(i);
                if (c == ' ' && !openQuote) {  // end of this parameter
                    // extract parameter
                    paramList.add(paramStr.substring(paramStart, i).trim());
                    // set start index
                    paramStart = i+1;
                } else if (c == '"') {
                    if (openQuote) {
                        openQuote = false;
                    } else {
                        openQuote = true;
                    }
                }
            }
            // account for the final param
            if (paramStart < i) {
                paramList.add(paramStr.substring(paramStart).trim());
            }
            // store param list as array
            stmt.setParams(paramList.toArray());
        } else {
            stmt.setParams(new Object[0]);
        }
        return stmt;
    }

    /**
     * Determines the type of the supplied parameter, and returns an
     * object representing it, e.g., int->Integer, double->Double, or String.
     * 
     * @param param  The parameter to parse for type
     * @return Object  the resulting object representing the basic type
     */
    public static Object parseParam (String param) {
        return parseParam(param, true);
    }

    public static Object parseParam (String param, boolean detType) {
        if (param == null) return null;

        Object valObj = param;

        if (param.startsWith("\"")) {  // this param is a string literal
            // strip the quotes
            valObj = param.substring(1, param.length()-1);
        } else {
            if (detType) {
                // test for primitive number types
                try {
                    valObj = Integer.valueOf(param);
                } catch (NumberFormatException e) {
                    // no, try double
                    try {
                        valObj = Double.valueOf(param);
                    } catch (NumberFormatException e1) {
                        // not double either, treat as variable and fetch value
                        valObj = getVar(param);
                        if (valObj == null) {
                            // treat as String literal
                            valObj = param;
                        }
                    }
                }
            }
        }
        
        return valObj;
    }

    public static void setVar (String name, Object val) {
        _varMap.put(name, val);
    }
    
    public static Object getVar (String name) {
        return _varMap.get(name);
    }

    public static void removeVar (String name) {
        _varMap.remove(name);
    }
    
    public static String[] listVars () {
        return (String[]) _varMap.keySet().toArray(new String[0]);
    }
    
    public static void clearVars () {
        _varMap.clear();
    }


    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append(cmd()).append(" ");  // command
        if (params().length > 0) {
            // parameters
            buf.append(Statement.parseParam((String) params()[0]));
            for (int i=1; i < params().length; i++) {
                buf.append(" ");
                Object val = Statement.parseParam((String) params()[i]);
                if (val instanceof String) {
                    buf.append("\"").append(val).append("\"");
                } else {
                    buf.append(val);
                }
            }
        }
        return buf.toString();
    }

    public boolean isLocal () {
        return StatementFactory.isLocalStatement(cmd());
    }

    /**
     * Takes a command statement and processes the method referred to in the
     * statement, using the parameters supplied, and invokes the method on
     * the remote interface via reflection.  Result is stored if a variable is
     * present.<br>
     * <br>
     * Statements take the form:<blockquote>
     *     <code>[var] [=] &lt;method&gt;(params...)</code>
     * @param stmt  the statement to process
     * @return boolean  whether method executed successfully
     */
    public boolean execute (Object server) {
        boolean rv = false;

        // see if change server
        if (server() != null) {  // yes
            server = server();
        }

        // invoke the method
        Class clazz = server.getClass();

        Object [] params = new Object[params().length];
        for (int i=0; i < params.length; i++) {  // get the true param value
            params[i] = Statement.parseParam((String) params()[i]);
        }

        if (_paramTypes == null) {  // param types haven't been processed yet
            // process the parameters
            _paramTypes = new Class[params.length];
            for (int i=0; i < params.length; i++) {
                if (params[i] instanceof Integer) {
                    _paramTypes[i] = int.class;
                } else if (params[i] instanceof Double) {
                    _paramTypes[i] = double.class;
                } else {
                    _paramTypes[i] = params[i].getClass();
                }
            }
        }

        try {
            System.out.println("Executing " + toString() + "...");
            Method method = clazz.getMethod(cmd(), _paramTypes);
            Object retVal = method.invoke(server, params);
            if (retVal != null) {
                if (var() != null) {
                    setVar(var(), retVal);
                    System.out.println("-> " + var() + " = (return value) " + retVal);
                } else {
                    if (retVal instanceof Remote) {
                        System.out.println("Return value is a remote interface, so it's stored in a variable.");
                        setVar(REMOTE_VAR, retVal);
                        System.out.println("-> " + REMOTE_VAR + " = (return value) " + retVal.toString());
                    } else {
                        System.out.println("-> Return value: " + retVal);
                    }
                }
            }
            rv = true;
        } catch (NoSuchMethodException e) {
            System.err.println("Method " + cmd() + " cannot be executed: " + e.getMessage());
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }

        return rv;
    }

    public String usage () {
        return "";
    }

    public String var () {
        return _retVar;
    }
    
    public void setReturnVar (String var) {
        _retVar = var;
    }
    
    public Remote server () {
        // get the Remote server object
        Remote server = null;
        try {
            server = (Remote) Statement.getVar(_server);
        } catch (ClassCastException e) {
            System.err.println("** Var '" + _server + "' does not refer to a Remote server object!");
            System.err.println("    " + e.getMessage());
            server = null;
        }

        return server;
    }
    
    public void setServer (String server) {
        _server = server;
    }
    
    public String cmd () {
        return _cmd;
    }
    
    public void setCmd (String cmd) {
        _cmd = cmd;
    }
    
    public Object[] params () {
        return _params;
    }
    
    public void setParams (Object[] params) {
        _params = params;
    }
}